"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
from typing import Optional, Union
import torch.nn as nn

from torchvision.models import ResNet
from torchvision.models.resnet import Bottleneck, BasicBlock, WeightsEnum
from torch import Tensor


class BackboneResnet(ResNet):
    """
    2D ResNet backbone
    """

    def __init__(
        self,
        *,
        pretrained: bool = False,
        weights: Optional[Union[WeightsEnum, dict]] = None,
        in_channels: int = 3,
        name: str = "resnet18",
        pool: Optional[str] = None,
    ) -> None:
        """
        Create 2D Resnet
        :param pretrained: reload imagenet weights
        :param in_channels: Number of input channels
        :param name: model name. Currently support 'resnet18' and 'resnet50'
        :param pool: whether to use global average pooling to reduce the spacial dimensions after the convolutional layers can be either 'avg', 'max' or None - dont use pool

        """
        # init parameters per required backbone
        init_parameters = {
            "resnet18": [BasicBlock, [2, 2, 2, 2]],
            "resnet34": [BasicBlock, [3, 4, 6, 3]],
            "resnet50": [Bottleneck, [3, 4, 6, 3]],
        }[name]
        # init original model
        super().__init__(*init_parameters)

        # load pretrained parameters if required
        if weights is not None and pretrained:
            raise Exception(
                "Use only one method to load pre-trained weights. Two were given!"
            )

        if pretrained:
            print(
                "Warning: not supported by new torchvision version - use weights instead"
            )
            from torch.hub import load_state_dict_from_url
            from torchvision.models.resnet import model_urls

            state_dict = load_state_dict_from_url(model_urls[name])
            self.load_state_dict(state_dict)

        if weights is not None:
            if isinstance(weights, WeightsEnum):
                self.load_state_dict(weights.get_state_dict(progress=True))
            elif isinstance(weights, dict):
                self.load_state_dict(weights, strict=False)

        del self.fc

        # save input parameters
        self.pretrained = pretrained
        self.in_channels = in_channels
        if pool == "avg":
            self.pool_layer = nn.AdaptiveAvgPool2d((1, 1))
        elif pool == "max":
            self.pool_layer = nn.AdaptiveMaxPool2d((1, 1))

        self._pool = pool
        if self.in_channels != 3:
            # modify the first convolution layer to support any number of input channels
            if self.in_channels == 1:
                self.conv1.in_channels = 1
                self.conv1.weight = nn.Parameter(
                    self.conv1.weight.sum(dim=1, keepdim=True)
                )
            else:
                self.conv1 = nn.Conv2d(
                    self.in_channels, 64, kernel_size=7, stride=2, padding=3, bias=False
                )

    def forward(self, x: Tensor) -> Tensor:  # type: ignore
        """
        Forward pass extract spatial features
        :param x: input tensor. Shape [batch_size, input_channels, y, x]
        :return: spatial features [batch_size, n_features, y', x']
        """
        # x assumed to be [batch_size, in_channels, y, x]
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)
        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)
        if hasattr(self, "_pool"):
            if self._pool is not None:
                x = self.pool_layer(x)
                x = x.flatten(1)
        return x



import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Tuple

# Based on the architecture from https://github.com/MrGiovanni/ModelsGenesis
torch.use_deterministic_algorithms(False)


class ContBatchNorm3d(nn.modules.batchnorm._BatchNorm):
    def _check_input_dim(self, input: torch.Tensor) -> None:

        if input.dim() != 5:
            raise ValueError("expected 5D input (got {}D input)".format(input.dim()))
        # super(ContBatchNorm3d, self)._check_input_dim(input)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        self._check_input_dim(input)
        return F.batch_norm(
            input,
            self.running_mean,
            self.running_var,
            self.weight,
            self.bias,
            True,
            self.momentum,
            self.eps,
        )


class LUConv(nn.Module):
    def __init__(self, in_chan: int, out_chan: int, act: str):
        super(LUConv, self).__init__()
        self.conv1 = nn.Conv3d(in_chan, out_chan, kernel_size=3, padding=1)
        self.bn1 = ContBatchNorm3d(out_chan)

        if act == "relu":
            self.activation = nn.ReLU(out_chan)
        elif act == "prelu":
            self.activation = nn.PReLU(out_chan)
        elif act == "elu":
            self.activation = nn.ELU(inplace=True)
        else:
            raise

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = self.activation(self.bn1(self.conv1(x)))
        return out


def _make_nConv(
    in_channel: int, depth: int, act: str, double_chnnel: bool = False
) -> nn.Module:
    if double_chnnel:
        layer1 = LUConv(in_channel, 32 * (2 ** (depth + 1)), act)
        layer2 = LUConv(32 * (2 ** (depth + 1)), 32 * (2 ** (depth + 1)), act)
    else:
        layer1 = LUConv(in_channel, 32 * (2**depth), act)
        layer2 = LUConv(32 * (2**depth), 32 * (2**depth) * 2, act)

    return nn.Sequential(layer1, layer2)


# class InputTransition(nn.Module):
#     def __init__(self, outChans, elu):
#         super(InputTransition, self).__init__()
#         self.conv1 = nn.Conv3d(1, 16, kernel_size=5, padding=2)
#         self.bn1 = ContBatchNorm3d(16)
#         self.relu1 = ELUCons(elu, 16)
#
#     def forward(self, x):
#         # do we want a PRELU here as well?
#         out = self.bn1(self.conv1(x))
#         # split input in to 16 channels
#         x16 = torch.cat((x, x, x, x, x, x, x, x,
#                          x, x, x, x, x, x, x, x), 1)
#         out = self.relu1(torch.add(out, x16))
#         return out


class DownTransition(nn.Module):
    def __init__(self, in_channel: int, depth: int, act: str):
        super(DownTransition, self).__init__()
        self.ops = _make_nConv(in_channel, depth, act)
        self.maxpool = nn.MaxPool3d(2)
        self.current_depth = depth

    def forward(self, x: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        if self.current_depth == 3:
            out = self.ops(x)
            out_before_pool = out
        else:
            out_before_pool = self.ops(x)
            out = self.maxpool(out_before_pool)
        return out, out_before_pool


class UpTransition(nn.Module):
    def __init__(self, inChans: int, outChans: int, depth: int, act: str):
        super(UpTransition, self).__init__()
        self.depth = depth
        self.up_conv = nn.ConvTranspose3d(inChans, outChans, kernel_size=2, stride=2)
        self.ops = _make_nConv(inChans + outChans // 2, depth, act, double_chnnel=True)

    def forward(self, x: torch.Tensor, skip_x: torch.Tensor) -> torch.Tensor:
        out_up_conv = self.up_conv(x)
        concat = torch.cat((out_up_conv, skip_x), 1)
        out = self.ops(concat)
        return out


class OutputTransition(nn.Module):
    def __init__(self, inChans: int, n_labels: int):

        super(OutputTransition, self).__init__()
        self.final_conv = nn.Conv3d(inChans, n_labels, kernel_size=1)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = self.sigmoid(self.final_conv(x))
        return out


class UNet3D(nn.Module):
    # the number of convolutions in each layer corresponds
    # to what is in the actual prototxt, not the intent
    def __init__(self, n_class: int = 1, act: str = "relu", for_cls: bool = False):
        super(UNet3D, self).__init__()
        self.for_cls = for_cls
        self.down_tr64 = DownTransition(1, 0, act)
        self.down_tr128 = DownTransition(64, 1, act)
        self.down_tr256 = DownTransition(128, 2, act)
        self.down_tr512 = DownTransition(256, 3, act)
        if not for_cls:
            self.up_tr256 = UpTransition(512, 512, 2, act)
            self.up_tr128 = UpTransition(256, 256, 1, act)
            self.up_tr64 = UpTransition(128, 128, 0, act)
            # self.out_tr = OutputTransition(64, n_class)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        self.out64, self.skip_out64 = self.down_tr64(x)
        self.out128, self.skip_out128 = self.down_tr128(self.out64)
        self.out256, self.skip_out256 = self.down_tr256(self.out128)
        self.out512, self.skip_out512 = self.down_tr512(self.out256)
        if self.for_cls:
            return self.out512

        self.out_up_256 = self.up_tr256(self.out512, self.skip_out256)
        self.out_up_128 = self.up_tr128(self.out_up_256, self.skip_out128)
        self.out_up_64 = self.up_tr64(self.out_up_128, self.skip_out64)
        # self.out = self.out_tr(self.out_up_64)

        return self.out_up_64



import hydra
from omegaconf import DictConfig, OmegaConf
from fuse.data.tokenizers.modular_tokenizer.modular_tokenizer import ModularTokenizer
from typing import Dict, Any


@hydra.main(
    config_path="../pretrained_tokenizers/configs",
    config_name="config_add_tokenizer_to_multi_tokenizer",
    version_base=None,
)
def main(cfg: DictConfig) -> None:
    """script to add a tokenizer (and all special tokens from it and special_tokens.py) to an existing tokenizer.
    The old tokenizer is read from the in_path, tokenizer to add is taken from the tokenizer_to_add variable.
    max_possible_token_id will be updated if the new max is larger then the old one.
    Add the tokenizer_info of the new tokenizer, as usual.


    Args:
        cfg (DictConfig): the config file.
    """

    cfg = hydra.utils.instantiate(cfg)
    tmp = OmegaConf.to_object(cfg)
    cfg_raw: Dict[str, Any] = tmp

    new_max_token_id = cfg_raw["data"]["tokenizer"]["max_possible_token_id"]

    t_mult = ModularTokenizer.load(path=cfg_raw["data"]["tokenizer"]["in_path"])

    # Update tokenizer with special tokens:
    new_tokenizer_name = cfg_raw["data"]["tokenizer"]["tokenizer_to_add"]
    cfg_tokenizer_info = {
        info["name"]: info for info in cfg_raw["data"]["tokenizer"]["tokenizers_info"]
    }
    new_tokenizer_info = cfg_tokenizer_info[new_tokenizer_name]
    if new_max_token_id > t_mult._max_possible_token_id:
        print(
            f"updating the max possible token ID from {t_mult._max_possible_token_id} to {new_max_token_id}"
        )
        t_mult._max_possible_token_id = new_max_token_id

    t_mult.add_single_tokenizer(new_tokenizer_info)
    t_mult.save(path=cfg_raw["data"]["tokenizer"]["out_path"])
    print("Fin")


if __name__ == "__main__":
    main()



from .common import ClassifierFCN, ClassifierFCN3D, ClassifierMLP
from .heads_1D import Head1D
from .heads_3D import Head3D
from .head_dense_segmentation import HeadDenseSegmentation
from .head_global_pooling_classifier import HeadGlobalPoolingClassifier



from typing import Optional, Sequence, Hashable, List, Dict, Callable, Any
from fuse.eval.metrics.libs.ensembling import Ensembling
from fuse.eval.metrics.utils import PerSampleData

import pandas as pd

import numpy as np

from fuse.eval.metrics.metrics_common import MetricDefault
from functools import partial


class MetricEnsemble(MetricDefault):
    def __init__(
        self,
        pred_keys: Sequence[str],
        target: Optional[str] = None,
        method: Optional[str] = "average",
        output_file: Optional[str] = None,
        output_pred_key: Optional[str] = "preds",
        output_target_key: Optional[str] = "target",
        rename_in_output: Optional[Dict[str, str]] = None,
        scores_normalize_func: Optional[Callable] = None,
        dropna: bool = False,
        **kwargs: dict,
    ):
        """
        Model ensembling metric.
        It obtains as input predictions from multiple models on the same set of samples
        and outputs a refined prediction using average, median or voting.
        :param pred_keys: List of key names (strings) for the multiple model prediction scores
        :param target: Optional key for target values. Only needed for optional saving the targets in an output file.
        :param method: Ensembling method. 'average' or 'voting'.
            if 'average', the predictions are assumed to be continuous (probabilities or regression output)
            if 'voting', the predictions are assumed to be class predictions (integers)
        :param output_file: Optional output filename
        :param output_pred_key: output key name for the predictions
        :param output_target_key: output key name for the target
        :param rename_in_output: renaming keys in the output (similar to output_pred_key and output_target that rename the columns in output file correspongin to pred and target keys).
                                  All additional keys (provided in **kwargs) will be added as columns to the output file.
                                  This optional param allows specifying for each additional input key the name of the corresponding column in the output file.
                                  It is similar to output_pred_key, and output_target_key which specify column names for pred and targte keys.
        :param scores_normalize_func: applied to each set of predictions / scores
        :param dropna: whether to drop samples with mising predictions
        """
        ensemble = partial(
            self._ensemble,
            method=method,
            output_file=output_file,
            output_pred_key=output_pred_key,
            output_target_key=output_target_key,
            keys_for_output=list(kwargs.keys()),
            rename_in_output=rename_in_output,
            scores_normalize_func=scores_normalize_func,
            dropna=dropna,
        )
        for i, key in enumerate(pred_keys):
            kwargs["pred" + str(i)] = key
        super().__init__(
            metric_func=ensemble, target=target, extract_ids=True, **kwargs
        )

    def _ensemble(
        self,
        ids: Sequence[Hashable],
        target: Optional[Sequence[np.ndarray]] = None,
        method: Optional[str] = "average",
        output_file: Optional[str] = None,
        output_pred_key: Optional[str] = "preds",
        output_target_key: Optional[str] = "target",
        keys_for_output: Optional[List[str]] = None,
        rename_in_output: Optional[Dict[str, str]] = None,
        scores_normalize_func: Callable = None,
        dropna: bool = False,
        **kwargs: dict,
    ) -> Dict[str, Any]:
        preds = [np.asarray(kwargs[k]) for k in kwargs if k.startswith("pred")]
        if dropna:
            nan_mat = np.isnan(np.concatenate(preds, axis=1))
            has_nan_arr = np.sum(nan_mat, axis=1) > 0
            ids = [id for (id, has_nan) in zip(ids, has_nan_arr) if not has_nan]
            preds = [pred[~has_nan_arr] for pred in preds]

        if scores_normalize_func is not None:
            preds = [scores_normalize_func(x) for x in preds]
        preds = list(np.stack(preds, axis=1))
        ensemble_preds = Ensembling.ensemble(preds=preds, method=method)
        # make sure to return the per-sample metric result for the relevant sample ids:
        per_sample_data = PerSampleData(data=ensemble_preds, ids=ids)
        res = {output_pred_key: per_sample_data}
        if target is not None:
            res[output_target_key] = PerSampleData(data=target, ids=ids)
        if keys_for_output is not None:
            for key in keys_for_output:
                key2 = rename_in_output.get(key, key)
                res[key2] = PerSampleData(data=kwargs[key], ids=ids)
        if output_file is not None:
            ids_list = list(ids)  # to fix the order of ids
            file_data = {"id": ids_list}
            for k, per_sample_data in res.items():
                file_data[k] = per_sample_data(ids_list)
            df = pd.DataFrame(file_data)

            df.to_pickle(output_file, compression="gzip")
        return res



# version
from fuse.version import __version__

# workaround for an issue caused by importing pandas after numpy
import pandas as pd

from fuse.utils.ndict import NDict
from fuse.utils.data.collate import CollateToBatchList, uncollate

from fuse.utils.rand.param_sampler import (
    Uniform,
    RandInt,
    RandBool,
    Choice,
    draw_samples_recursively,
)
from fuse.utils.rand.seed import Seed

set_seed = Seed.set_seed
from fuse.utils.file_io.file_io import read_dataframe, save_dataframe
from fuse.utils.data.collate import CollateToBatchList, uncollate



# type: ignore
class DummyContext:
    def __init__(self, *args, **kwargs) -> None:
        pass

    def __enter__(self, *args, **kwargs):
        pass

    def __exit__(self, *args, **kwargs):
        pass



import torch
from torch import nn
from typing import Optional
from vit_pytorch.vit import Transformer as _Transformer
from vit_pytorch.vit import repeat
from x_transformers import Encoder, CrossAttender, TransformerWrapper


class Transformer(nn.Module):
    """
    Transformer backbone.
    Gets a [batch_size, num_tokens, token_dim] shaped tensor
    Returns a [batch_size, num_tokens + num_cls_tokens, token_dim] shaped tensor, where the first tokens are the CLS tokens
    """

    def __init__(
        self,
        *,
        num_tokens: int,
        token_dim: int,
        depth: int,
        heads: int,
        mlp_dim: int,
        dim_head: int = 64,
        dropout: float = 0.0,
        emb_dropout: float = 0.0,
        num_cls_tokens: int = 1
    ):
        super().__init__()
        self.num_cls_tokens = num_cls_tokens
        self.pos_embedding = nn.Parameter(
            torch.randn(1, num_tokens + num_cls_tokens, token_dim)
        )
        self.cls_tokens = nn.Parameter(torch.randn(1, num_cls_tokens, token_dim))
        self.dropout = nn.Dropout(emb_dropout)
        self.transformer = _Transformer(
            dim=token_dim,
            depth=depth,
            heads=heads,
            dim_head=dim_head,
            mlp_dim=mlp_dim,
            dropout=dropout,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        :param x: [batch_size, num_tokens, token_dim] shaped tensor
        :return: [batch_size, num_tokens + num_cls_tokens, token_dim] shaped tensor, where the first tokens are the CLS tokens
        """
        b, n, _ = x.shape
        cls_tokens = repeat(self.cls_tokens, "1 a d -> b a d", b=b)
        x = torch.cat((cls_tokens, x), dim=1)
        x += self.pos_embedding
        x = self.dropout(x)
        x = self.transformer(x)
        return x


class CrossAttentionTransformerEncoder(nn.Module):
    """
    CrossAttentionTransformerEncoder backbone model based on x-transformers library.

    Input:
        two sequences 'seq_a, seq_b' with shapes [batch_size, len(seq_a)], [batch_size, len(seq_b)] respectively.
    Output:
        features tensor with shape [batch_size, output_dim]


    Architecture:
        The model consist of the following blocks:
            * Two encoding layers - one for each sequence
            * One or two cross attention (1) layers - depends on the user's request
            * One linear layer

    Forward Pass:
        -> Receive two sequences as inputs, assuming both of them are already tokenized.
        -> Pass each of the sequences through it's own encoder (extracting features)
        -> Performs cross attention with the two encoded features using one of the sequences as context
                (also supports using both as sequence:
                    * performs two times cross attention, each time using a different sequence as the context
                    * concat both outputs into one vector)
        -> Apply linear layer on the last encoded vector

    Building Blocks:
        In this architecture we use three components from "x-transformers" library.
        Here is a short summary - for more information consider check source code.
        * TransformerWrapper - wraps an attention layer (in our case the Encoder) and applies token & positional embedding.
        * Encoder - self attention layers. In our case it gets the embedding from the wrapper.
        * CrossAttender - cross attention layers. In our case it gets the embedding from the encoders.

    (1) see the following blog post for more info regarding cross attention in transformers:
        https://vaclavkosar.com/ml/cross-attention-in-transformer-architecture
    """

    def __init__(
        self,
        emb_dim: int,
        num_tokens_a: int,
        num_tokens_b: int,
        max_seq_len_a: int,
        max_seq_len_b: int,
        depth_a: int = 6,
        depth_b: int = 6,
        depth_cross_attn: int = 6,
        heads_a: int = 9,
        heads_b: int = 9,
        output_dim: Optional[int] = None,
        context: str = "seq_b",
        kwargs_wrapper_a: Optional[dict] = None,
        kwargs_wrapper_b: Optional[dict] = None,
        kwargs_encoder_a: Optional[dict] = None,
        kwargs_encoder_b: Optional[dict] = None,
        kwargs_cross_attn: Optional[dict] = None,
    ):
        """
        :param emb_dim: inner model dimension
        :param num_tokens_a: number of tokens of the first sequence
        :param num_tokens_b: number of tokens of the second sequence
        :param max_seq_len_a: the maximum length of the first sequence
        :param max_seq_len_b: the maximum length of the second sequence
        :param depth_a: first sequence encoder's depth
        :param depth_b: second sequence encoder's depth
        :param depth_cross_attn: cross attender(s)' length
        :param heads_a: number of attention heads for the first sequence's encoder
        :param heads_b: number of attention heads for the second sequence's encoder
        :param output_dim: (optional) model's output dimension. if not give the emb dim will be used as default.
        :param context: which sequence will be used as context in the cross attention module:
                        "seq_a": the first sequence will be used as a context
                        "seq_b": the second sequence will be used as a context
                        "both": will use two cross attention modules to take each one of the sequences as a context to the other one.
        :param kwargs_wrapper_a: optional - additional arguments for sequence a's TransformerWrapper object
        :param kwargs_wrapper_b: optional - additional arguments for sequence b's TransformerWrapper object
        :param kwargs_encoder_a: optional - additional arguments for sequence a's Encoder object
        :param kwargs_encoder_b: optional - additional arguments for sequence b's Encoder object
        :param kwargs_cross_attn: optional - additional arguments for the CrossAttender object(s)
        """
        super().__init__()

        if output_dim is None:
            output_dim = emb_dim

        assert context in ["seq_a", "seq_b", "both"]
        self._context = context

        # init sequences' encoders
        self.enc_a = TransformerWrapper(
            num_tokens=num_tokens_a,
            max_seq_len=max_seq_len_a,
            **kwargs_wrapper_a if kwargs_wrapper_a else dict(),
            attn_layers=Encoder(
                dim=emb_dim,
                depth=depth_a,
                heads=heads_a,
                **kwargs_encoder_a if kwargs_encoder_a else dict(),
            ),
        )
        self.enc_b = TransformerWrapper(
            num_tokens=num_tokens_b,
            max_seq_len=max_seq_len_b,
            **kwargs_wrapper_b if kwargs_wrapper_b else dict(),
            attn_layers=Encoder(
                dim=emb_dim,
                depth=depth_b,
                heads=heads_b,
                **kwargs_encoder_b if kwargs_encoder_b else dict(),
            ),
        )

        # cross attention module(s)
        if self._context in ["seq_a", "seq_b"]:
            self.cross_attn = CrossAttender(
                dim=emb_dim,
                depth=depth_cross_attn,
                **kwargs_cross_attn if kwargs_cross_attn else dict(),
            )

        else:  # both
            self.cross_attn_a_as_context = CrossAttender(
                dim=emb_dim,
                depth=depth_cross_attn,
                **kwargs_cross_attn if kwargs_cross_attn else dict(),
            )
            self.cross_attn_b_as_context = CrossAttender(
                dim=emb_dim,
                depth=depth_cross_attn,
                **kwargs_cross_attn if kwargs_cross_attn else dict(),
            )

        self.last_linear = nn.Linear(emb_dim, output_dim)

    def forward(self, xa: torch.Tensor, xb: torch.Tensor) -> torch.Tensor:
        """
        assumes input sequences are already tokenized

        :param xa: tensor with shape [batch_size, seq_len_a]
        :param xb: tensor with shape [batch_size, seq_len_b]
        :return: raw embeddings
        """
        # encoding stage
        enc_xa = self.enc_a(xa, return_embeddings=True)
        enc_xb = self.enc_b(xb, return_embeddings=True)

        # cross attention stage
        if self._context == "seq_a":
            x = self.cross_attn(enc_xb, context=enc_xa)

        elif self._context == "seq_b":
            x = self.cross_attn(enc_xa, context=enc_xb)

        else:
            x_acb = self.cross_attn_b_as_context(enc_xa, context=enc_xb)
            x_bca = self.cross_attn_a_as_context(enc_xb, context=enc_xa)
            x = torch.cat((x_acb, x_bca), dim=1)

        # linear layer
        x = self.last_linear(x)
        return x



import pickle
from collections import OrderedDict
from random import randrange
from typing import Sequence, Any, Tuple, List

import glob
import os
import numpy as np
import pandas as pd

from fuse.data import DatasetDefault, PipelineDefault, OpBase
from fuse.data.ops.ops_read import OpReadDataframe
from fuse.data.ops.ops_common import OpLookup, OpSetIfNotExist
from fuse.utils.ndict import NDict

from fuse.data.utils.split import dataset_balanced_division_to_folds
from fuse.data.utils.export import ExportDataset

from fuse_examples.multimodality.ehr_transformer.utils import (
    seq_translate,
    position_idx,
    special_tokens,
    seq_pad,
    WordVocab,
)


VALID_TESTS_ABOVE_ZERO = [
    "pH",
    "Weight",
    "Height",
    "DiasABP",
    "HR",
    "NIMAP",
    "MAP",
    "NIDiasABP",
    "NISysABP",
    "PaCO2",
    "PaO2",
    "Temp",
    "SaO2",
    "RespRate",
    "SysABP",
]

STATIC_FIELDS = ["Age", "Gender", "Height", "ICUType", "Weight"]


class OpAddBMI(OpBase):

    """
    Apply BMI calculation using static patient's fields. Names of the fields
    are provided as parameters as well as output field name.
    Height expected to be in centimeters
    Weight expected to be in kilograms

    Example:
    OpAddBMI(key_in_height="Height", key_in_weight="Weight", key_out_bmi="BMI")
    """

    def __init__(self, key_in_height: str, key_in_weight: str, key_out_bmi: str):
        super().__init__()
        self._key_height = key_in_height
        self._key_weight = key_in_weight
        self._key_bmi = key_out_bmi

    def __call__(self, sample_dict: NDict) -> Any:
        if (self._key_height in sample_dict["StaticDetails"]) & (
            self._key_weight in sample_dict["StaticDetails"]
        ):
            height = sample_dict["StaticDetails." + self._key_height]
            weight = sample_dict["StaticDetails." + self._key_weight]
            if ~np.isnan(height) & ~np.isnan(weight):
                sample_dict["StaticDetails." + self._key_bmi] = (
                    10000 * weight / (height * height)
                )

        return sample_dict


class OpConvertVisitToSentence(OpBase):
    """
    Converts the observations presented in data frame
     to sentence by visit time. Sentence will be presented as
     list of values per visit (converted to categorical strings)
     and separator at the end of the sentence
    """

    def __call__(self, sample_dict: NDict) -> Any:
        df_visits = sample_dict["Visits"]

        d_visit_sentences = OrderedDict()

        for visit_time, df_visit in df_visits.groupby("Time", sort=True):
            d_visit_sentences[visit_time] = []

            d_visit_sentences[visit_time].extend(df_visit["Value"].to_list())
            d_visit_sentences[visit_time].extend(["SEP"])

        sample_dict["VisitSentences"] = d_visit_sentences

        return sample_dict


class OpGenerateFinalTrajectoryOfVisits(OpBase):
    """
    Generates final trajectory of visits before training including:
        - random extraction of sub-trajectory (as augmentation procedure)
        - limitation its size by configured max_len of a sequence
        - embedding static features, according to defined configuration
        - adding special tokens of classifier and padding to the final sequence
        - translating tokens to indexes
        - computation of positional indexes of sentences for further embedding
    """

    def __init__(
        self,
        max_len: int,
        vocab: dict,
        embed_static_in_all_visits: int,
        static_variables_to_embed: list,
    ):
        super().__init__()
        self._max_len = max_len
        self._vocab = vocab
        self._static_variables_to_embed = static_variables_to_embed
        self._embed_static_in_all_visits = embed_static_in_all_visits

    def __call__(self, sample_dict: NDict) -> Any:
        d_visits_sentences = sample_dict["VisitSentences"]
        n_visits = len(d_visits_sentences)

        # extract sub-set of trajectory by random selection of
        # start and stop keys we use n_visits- 1 to keep the last
        # visit for outcome in classifier
        start_visit = randrange(int((n_visits - 1) / 2))
        stop_visit = n_visits - randrange(int((n_visits - 1) / 2)) - 1
        keys = list(d_visits_sentences.keys())
        trajectory_keys = keys[start_visit:stop_visit]
        next_visit = keys[stop_visit]

        # add static variables as a first special visit
        df_static = sample_dict["StaticDetails"]
        static_embeddings = []
        if len(self._static_variables_to_embed) > 0:
            for k in df_static.keys():
                if k in self._static_variables_to_embed:
                    static_embeddings += [df_static[k]]

        # Build trajectory of visits
        # appends sentences in the reverse order from
        # stop to start in order to get X last sentences that
        # less than max_len
        trajectory_tokens = []
        for k in reversed(trajectory_keys):
            visit_sentence = d_visits_sentences[k]

            # add static details to every visit
            if self._embed_static_in_all_visits:
                visit_sentence = static_embeddings + visit_sentence

            if (len(trajectory_tokens) + len(visit_sentence)) < self._max_len:
                trajectory_tokens = visit_sentence + trajectory_tokens
            else:
                break

        # create final sequence of tokens
        tokens = [special_tokens["cls"]]
        if not self._embed_static_in_all_visits:
            static_embeddings += [special_tokens["separator_static"]]
            tokens += static_embeddings
        tokens += trajectory_tokens

        tokens = seq_pad(tokens, self._max_len)
        positions = position_idx(tokens)
        token_ids = seq_translate(tokens, self._vocab)

        sample_dict["Tokens"] = tokens
        sample_dict["Positions"] = positions
        sample_dict["Indexes"] = np.array(token_ids).squeeze(0)

        # outcomes of next visit prediction
        sample_dict["NextVisitTokens"] = d_visits_sentences[next_visit]
        sample_dict["NextVisitTokenIds"] = seq_translate(
            sample_dict["NextVisitTokens"], self._vocab
        )
        sample_dict["NextVisitLabels"] = np.zeros(len(self._vocab))
        sample_dict["NextVisitLabels"][sample_dict["NextVisitTokenIds"]] = 1.0

        return sample_dict


class OpMapToCategorical(OpBase):
    """
    Converts continuous values of observations
    to the category based on calculated "bins"
    each pair of name + value will be converted
    to the word "Name_binN"

     Example: Age = 80 will be converted to "Age_3"
      (third bin in calculated distribution)
    """

    def __call__(self, sample_dict: NDict, bins: dict) -> Any:
        # convert continuous measurements to categorical ones based
        # on defined bins mapping static clinical characteristics
        # (Age, Gender, ICU type, Height, etc)
        for k in sample_dict["StaticDetails"]:
            sample_dict[f"StaticDetails.{k}"] = (
                k + "_" + str(np.digitize(sample_dict["StaticDetails"][k], bins[k]))
            )

        # mapping labs exams and clinical characteristics captured
        # during patients' stay in ICU
        if not sample_dict["Visits"].empty:
            sample_dict["Visits"]["Value"] = sample_dict["Visits"].apply(
                lambda row: row["Parameter"]
                + "_"
                + str(np.digitize(row["Value"], bins[row["Parameter"]]))
                if row["Parameter"] in bins
                else special_tokens["unknown"],
                axis=1,
            )

        return sample_dict


class PhysioNetCinC:
    @staticmethod
    def _read_raw_data(raw_data_path: str) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Loads raw data from the provided directory path.
         It includes clinical patients data and outcomes
        :param raw_data_path: path to folder with raw data
        :return: data frame of patients observation records
        (record per observation) and data frame of outcomes
         (record per patient)
        """
        # read patients info and lab exams
        df = pd.DataFrame(columns=["PatientId", "Time", "Parameter", "Value"])
        data_sub_sets = ["set-a", "set-b"]
        for s in data_sub_sets:
            csv_files = glob.glob(os.path.join(raw_data_path + "/" + s, "*.txt"))
            for f in csv_files:
                patient_id = os.path.splitext(os.path.basename(f))[0]
                df_file = pd.read_csv(f)
                df_file = df_file.drop(
                    df_file[(df_file["Parameter"] == "RecordID")].index
                ).reset_index(drop=True)
                df_file["PatientId"] = patient_id
                df = pd.concat([df, df_file], axis=0)
        df.reset_index(inplace=True, drop=True)
        patient_ids = np.unique(df["PatientId"].values)

        # read outcomes
        df_outcomes = pd.DataFrame(columns=["RecordID", "In-hospital_death"])
        outcomes = ["Outcomes-a.txt", "Outcomes-b.txt"]
        for o in outcomes:
            o_file = os.path.join(raw_data_path + "/" + o)
            df_outcomes = pd.concat(
                [df_outcomes, pd.read_csv(o_file)[["RecordID", "In-hospital_death"]]]
            ).reset_index(drop=True)

        df_outcomes["RecordID"] = df_outcomes["RecordID"].astype(str)
        df_outcomes.rename(columns={"RecordID": "PatientId"}, inplace=True)

        # synchronize with patients data
        df_outcomes = df_outcomes[df_outcomes["PatientId"].isin(patient_ids)]

        return df, df_outcomes

    @staticmethod
    def _drop_records_with_errors(df: pd.DataFrame) -> pd.DataFrame:
        """
        Drop records with illegal values of observations
        :param df: data frame with observational records
        :return: updated data frame of patients' observations
        """
        # drop records with measurements below or equal
        # to zero for tests with only positive values
        for v in VALID_TESTS_ABOVE_ZERO:
            df = df.drop(
                df[(df["Parameter"] == v) & (df["Value"] <= 0)].index
            ).reset_index(drop=True)

        # drop gender values below zero
        df = df.drop(
            df[(df["Parameter"] == "Gender") & (df["Value"] < 0)].index
        ).reset_index(drop=True)

        # drop records with invalid Height and Weight values
        df = df.drop(
            df[(df["Parameter"] == "Weight") & (df["Value"] < 20)].index
        ).reset_index(drop=True)
        df = df.drop(
            df[(df["Parameter"] == "Height") & (df["Value"] < 100)].index
        ).reset_index(drop=True)
        return df

    @staticmethod
    def _drop_bad_patients(
        df: pd.DataFrame, min_hours: int, min_num_of_visits: int
    ) -> pd.DataFrame:
        """
        Drop records of patients that have the less than
        min_hours reported in ICU and  patient that have less
         than minimum number of visits
        :param df:
        :param min_hours: minimum number of hours of taken observations
        :param min_num_of_visits: minimum number of
        visits (observations taking events)
        :return: updated data frame of patients' observation records
        """

        df_fixed = df.copy()
        count_dropped_short_time = 0
        count_dropped_min_visits = 0
        for pat_id, df_pat_records in df.groupby("PatientId"):
            hours = df_pat_records["Time"].str.split(":", n=1, expand=True)[0].values
            if max(hours.astype(int)) < min_hours:
                df_fixed.drop(df_pat_records.index, inplace=True)
                count_dropped_short_time += 1
                continue

            num_visits = len(np.unique(df_pat_records["Time"].values))
            if num_visits < min_num_of_visits:
                df_fixed.drop(df_pat_records.index, inplace=True)
                count_dropped_min_visits += 1

        return df_fixed

    @staticmethod
    def _combine_data_by_patients(
        df_records: pd.DataFrame, df_outcomes: pd.DataFrame
    ) -> pd.DataFrame:
        """
        Combine all available information : static details,
        visits and target outcome in one data frame of patients
        :param df_records: data frame of observation records
        :param df_outcomes: date frame of patients' outcomes
        :return: combined data frame of patients
        """

        static_fields = ["Age", "Height", "Weight", "Gender", "ICUType"]

        patient_ids = []
        static_details = []
        visits = []
        target = []
        for pat_id, df_pat_records in df_records.groupby("PatientId"):
            patient_ids += [pat_id]
            # extract static patients details
            static_info_ind = (df_pat_records["Time"] == "00:00") & (
                df_pat_records["Parameter"].isin(static_fields)
            )
            df_pat_static = df_pat_records[static_info_ind]
            static_details += [dict(zip(df_pat_static.Parameter, df_pat_static.Value))]

            # extract visits (time series events)
            df_pat_dynamic_exams = df_pat_records[~static_info_ind]
            visits += [df_pat_dynamic_exams.copy()]

            label = df_outcomes[df_outcomes["PatientId"] == pat_id][
                "In-hospital_death"
            ].values[0]
            target += [label]

        list_of_tuples = list(zip(patient_ids, static_details, visits, target))
        df_patients = pd.DataFrame(
            list_of_tuples, columns=["PatientId", "StaticDetails", "Visits", "Target"]
        )

        return df_patients

    @staticmethod
    def _generate_percentiles(
        dataset: DatasetDefault,
        num_percentiles: int,
        categorical_max_num_of_values: int,
    ) -> dict:
        """

        :param dataset: data frame of observational records
        :param num_percentiles: number of bins to digitize the
         continuous observations
        :param categorical_max_num_of_values: which observations can be
        counted as categorical based on number
        of unique values.
        :return: dictionary of calculated percentiles (bins)
        """

        df = ExportDataset.export_to_dataframe(
            dataset, keys=["StaticDetails", "Visits"], workers=1
        )

        # Extracting static and dynamic parts of the dataset
        df_static = pd.DataFrame(df["StaticDetails"].to_list())
        df_visits = pd.concat(df["Visits"].values)

        # generation dictionaries of values for static
        # and dynamic variables of dataset patients
        d_static = df_static.to_dict("list")
        d_visits = dict.fromkeys(np.unique(df_visits[["Parameter"]]))
        for k in d_visits.keys():
            d_visits[k] = df_visits[df_visits["Parameter"] == k]["Value"]

        d_all_values = d_visits
        d_all_values.update(d_static)

        # calculate percentiles for categorical parameters
        # (Gender, etc) update percentiles according to categories
        d_percentile = dict()
        percentiles = range(
            0, 100 + int(100 / num_percentiles), int(100 / num_percentiles)
        )
        for k in d_all_values.keys():
            values = np.array(d_all_values[k])
            values = values[~np.isnan(values)]
            # check number of unique values
            unique_values = set(values)
            if len(unique_values) < categorical_max_num_of_values:
                # categorical value
                unique_values = sorted(unique_values)
                # incrementing in 1 is needed for getting bin number
                # corresponding to the variable value
                # e.g. gender values will remain the same
                unique_values = [x + 1 for x in unique_values]
                d_percentile[k] = sorted(unique_values)
            else:
                d_percentile[k] = np.percentile(values, percentiles)

        return d_percentile

    @staticmethod
    def _build_corpus_of_words(dict_percentiles: dict) -> list:
        """
        Generates a list of possible words that will be generated
         by mapping to bins. This list will define a corpus to the
          final vocabulary
        :param dict_percentiles: dictionary with calculated
        percentiles (bins) for each observation type
        :return: corpus of words
        """

        corpus = list(special_tokens.values())

        for k in dict_percentiles.keys():
            num_bins = len(dict_percentiles[k])
            # Create words for all possible bins in percentile
            for b in range(0, num_bins + 1):
                word = k + "_" + str(b)
                corpus += [word]

        return corpus

    @staticmethod
    def _load_and_process_df(
        raw_data_path: str,
        raw_data_pkl: str,
        min_hours_in_hospital: int,
        min_number_of_visits: int,
    ) -> Tuple[pd.DataFrame, list]:
        """
        Applies all stages for generation patients' data frame.
        Manages generated data frame in pickle file.

        :param raw_data_path: path to the raw data folder
        :param raw_data_pkl: path to the pickle file with generated
        patients data frame
        :param min_hours_in_hospital: used for filtering illegal patients
        :param min_number_of_visits: used for filtering illegal patients
        :return: data frame of patients and list of patient ids
        """
        # if pickle available
        try:
            df_patients, patient_ids = pickle.load(open(raw_data_pkl, "rb"))

        except Exception:
            df_raw_data, df_outcomes = PhysioNetCinC._read_raw_data(raw_data_path)

            # drop records with invalid tests results
            df_raw_data = PhysioNetCinC._drop_records_with_errors(df_raw_data)

            # drop patients with less than minimum hours in hospital and less
            # than minimum number of visits
            df_raw_data = PhysioNetCinC._drop_bad_patients(
                df_raw_data, min_hours_in_hospital, min_number_of_visits
            )
            patient_ids = np.unique(df_raw_data["PatientId"].values)

            df_patients = PhysioNetCinC._combine_data_by_patients(
                df_raw_data, df_outcomes
            )

            if raw_data_pkl is not None:
                with open(raw_data_pkl, "wb") as f:
                    pickle.dump([df_patients, patient_ids], f)

        return df_patients, patient_ids

    @staticmethod
    def _process_dynamic_pipeline(
        dict_percentiles: dict,
        static_variables_to_embed: Any,
        embed_static_in_all_visits: Any,
        token2idx: Any,
        max_len: int,
    ) -> List[OpBase]:
        """
        Generates dynamic pipeline by stacking a list of Operators.
        :param dict_percentiles: percentiles of observations
        :param static_variables_to_embed: the list of static variables
        to embed in patients trajectory
        :param embed_static_in_all_visits: the way of embedding static
        variables
        :param token2idx: vocabulary of tokens to indexes
        :param max_len: maximal lengths of tokens' sequence
        :return: list of Operators
        """
        return [
            (OpMapToCategorical(), dict(bins=dict_percentiles)),
            (OpConvertVisitToSentence(), dict()),
            (
                OpGenerateFinalTrajectoryOfVisits(
                    max_len,
                    token2idx,
                    embed_static_in_all_visits,
                    static_variables_to_embed,
                ),
                dict(),
            ),
            (OpSetIfNotExist(), dict(key="StaticDetails.Gender", value=-1)),
            (
                OpLookup(map={"Gender_0": 0, "Gender_1": 1, -1: -1}),
                dict(key_in="StaticDetails.Gender", key_out="Gender"),
            ),
        ]

    @staticmethod
    def dataset(
        raw_data_path: str,
        raw_data_pkl: str,
        num_folds: int,
        split_filename: str,
        seed: int,
        reset_split: bool,
        train_folds: Sequence[int],
        validation_folds: Sequence[int],
        test_folds: Sequence[int],
        num_percentiles: int,
        categorical_max_num_of_values: int,
        min_hours_in_hospital: int,
        min_number_of_visits: int,
        static_variables_to_embed: Sequence[str],
        embed_static_in_all_visits: int,
        max_len_seq: int,
    ) -> Tuple[Any, DatasetDefault, DatasetDefault, DatasetDefault]:
        """
        This is the main methods for dataset generation. It includes the stages
         - raw data loading and processing for generation data frame with
         patients' records
         - preparation percentiles for patients observations mapping (on train)
         - split to train/validation/test
         - creation of dynamic pipline in two stages:
            first stage (before percentile calculation) will include loading
            data to sample-dicts and calculation additional observations (BMI)
            second stage (after percentiles calculation) will include mapping
            to categorical using percentiles as bins; converting observations
            to sentences and generation random sub-trajectories using this
            sentences, static details will be embedded in trajectories.
         - generation train, validation and test datasets with defined
            dynamic pipline

        below input parameters are defined in config,yaml in dataset group

        :param raw_data_path: path to raw data folder
        :param raw_data_pkl: path to the pickle file with loaded data
        :param num_folds: number of folds
        :param split_filename: file of splits
        :param seed: random seed
        :param reset_split: bool for split reset
        :param train_folds: indexes of train folds
        :param validation_folds: indexes of validation folds
        :param test_folds: indexes of test folds
        :param num_percentiles: number of percentiles for mapping continuous
         observations
        :param categorical_max_num_of_values: definition of categorical
        in observations
        :param min_hours_in_hospital: minimum allowed numbers of hours
        in hospital
        :param min_number_of_visits: minimum allowed numbers of visits
        :param static_variables_to_embed: list of static variables for
        embedding
        :param embed_static_in_all_visits: way of embedding static  details
        :param max_len_seq: maximal length of token sequence
        :return: vocabulary of tokens to indexes, datasets of train,
        validation and test
        """
        assert raw_data_path is not None

        df_patients, patient_ids = PhysioNetCinC._load_and_process_df(
            raw_data_path, raw_data_pkl, min_hours_in_hospital, min_number_of_visits
        )

        # first step of the pipline is to read data and add
        # additional features (e.g. BMI)
        dynamic_pipeline_ops = [
            (OpReadDataframe(df_patients, key_column="PatientId"), {}),
            (
                OpAddBMI(
                    key_in_height="Height", key_in_weight="Weight", key_out_bmi="BMI"
                ),
                dict(),
            ),
        ]
        dynamic_pipeline = PipelineDefault("cinc_dynamic", dynamic_pipeline_ops)

        dataset_all = DatasetDefault(patient_ids, dynamic_pipeline)
        dataset_all.create()

        folds = dataset_balanced_division_to_folds(
            dataset=dataset_all,
            output_split_filename=split_filename,
            keys_to_balance=["Target"],
            nfolds=num_folds,
            seed=seed,
            reset_split=reset_split,
            workers=1,
        )

        train_sample_ids = []
        for fold in train_folds:
            train_sample_ids += folds[fold]
        dataset_train = DatasetDefault(train_sample_ids, dynamic_pipeline)
        dataset_train.create()

        # calculate statistics of train set only and generate dictionary
        # of percentiles for mapping lab results to categorical for train,
        # validation and test
        dict_percentiles = PhysioNetCinC._generate_percentiles(
            dataset_train, num_percentiles, categorical_max_num_of_values
        )

        # generation corpus and tokenizer
        corpus = PhysioNetCinC._build_corpus_of_words(dict_percentiles)
        token2idx = WordVocab(corpus, max_size=None, min_freq=1).get_stoi()

        # update pipeline with Op using calculated percentiles
        dynamic_pipeline_ops = dynamic_pipeline_ops + [
            *PhysioNetCinC._process_dynamic_pipeline(
                dict_percentiles,
                static_variables_to_embed,
                embed_static_in_all_visits,
                token2idx,
                max_len_seq,
            )
        ]
        dynamic_pipeline = PipelineDefault("cinc_dynamic", dynamic_pipeline_ops)

        # update dynamic pipeline after calculating general statistics
        # and percentiles based on train dataset
        dataset_train._dynamic_pipeline = dynamic_pipeline

        validation_sample_ids = []
        for fold in validation_folds:
            validation_sample_ids += folds[fold]
        dataset_validation = DatasetDefault(validation_sample_ids, dynamic_pipeline)
        dataset_validation.create()

        test_sample_ids = []
        for fold in test_folds:
            test_sample_ids += folds[fold]
        dataset_test = DatasetDefault(test_sample_ids, dynamic_pipeline)
        dataset_test.create()

        return token2idx, dataset_train, dataset_validation, dataset_test


if __name__ == "__main__":
    token2idx, ds_train, ds_valid, ds_test = PhysioNetCinC.dataset(
        os.environ["CINC_DATA_PATH"],
        None,
        5,
        None,
        1234,
        True,
        [0, 1, 2],
        [3],
        [4],
        4,
        5,
        46,
        10,
        ["Age", "Gender", "BMI"],
        0,
        350,
    )



from fuse.data.datasets.dataset_default import DatasetDefault
from fuse.data.datasets.caching.samples_cacher import SamplesCacher
from fuse.data import PipelineDefault, OpToTensor
from fuse.data.ops.ops_common import OpLambda
from fuseimg.data.ops.image_loader import OpLoadImage
from fuse.data.ops.ops_read import OpReadDataframe

from functools import partial
from typing import Hashable, Optional, Sequence, Union, Tuple
import torch
import pandas as pd
from fuse_examples.imaging.oai_example.data.data_ops import (
    OpNormalizeMRI,
    OpResize3D,
    OpRandomFlip,
    OpSegToOneHot,
    OpRandomCrop,
)


class SegOAI:
    @staticmethod
    def sample_ids(df: pd.DataFrame) -> list:
        return SegOAI.get_existing_sample_ids(df)

    @staticmethod
    def get_existing_sample_ids(df: pd.DataFrame) -> list:
        """
        get all the sample ids that have a zip file in the specified path
        """
        existing_files = df["idx"].values
        return existing_files

    @staticmethod
    def static_pipeline(df: pd.DataFrame) -> PipelineDefault:
        static_pipeline = PipelineDefault(
            "static",
            [
                # decoding sample ID
                (OpReadDataframe(df, key_column="idx"), dict()),
                (OpLoadImage(""), dict(key_in="img_path", key_out="img", format="nib")),
                (OpLoadImage(""), dict(key_in="seg_path", key_out="seg", format="nib")),
                (OpNormalizeMRI(), dict(key="img", to_range=[0, 1])),
            ],
        )
        return static_pipeline

    @staticmethod
    def dynamic_pipeline(
        validation: bool = False, resize_to: Tuple[int, int, int] = [40, 224, 224]
    ) -> PipelineDefault:
        """
        Get suggested dynamic pipeline. including pre-processing that might be modified and augmentation operations.
        :param train : True iff we request dataset for train purpouse
        """

        dynamic_pipeline = PipelineDefault("dynamic", [])
        # augmentation

        if validation:
            if resize_to is not None:
                dynamic_pipeline.extend(
                    [
                        (OpResize3D(), dict(key="img", shape=resize_to)),
                        (
                            OpResize3D(),
                            dict(key="seg", shape=resize_to, segmentation=True),
                        ),
                    ]
                )
            else:
                print("no Resize")
        else:
            dynamic_pipeline.extend(
                [
                    (OpRandomCrop(), dict(key=["img", "seg"], scale=[0.7, 1.0])),
                    (OpResize3D(), dict(key="img", shape=resize_to)),
                    (OpResize3D(), dict(key="seg", shape=resize_to, segmentation=True)),
                    (OpRandomFlip(), dict(key=["img", "seg"])),
                ]
            )
        dynamic_pipeline.extend(
            [
                (OpToTensor(), dict(key="img", dtype=torch.float32)),
                (OpToTensor(), dict(key="seg", dtype=torch.int8)),
                (OpLambda(partial(torch.unsqueeze, dim=0)), dict(key="img")),
                (OpLambda(partial(torch.unsqueeze, dim=0)), dict(key="seg")),
                (OpSegToOneHot(n_classes=7), dict(key="seg")),
            ]
        )

        return dynamic_pipeline

    @staticmethod
    def dataset(
        csv_path: Union[str, pd.DataFrame],
        cache_dir: str = None,
        reset_cache: bool = True,
        num_workers: int = 10,
        sample_ids: Optional[Sequence[Hashable]] = None,
        resize_to: tuple = (40, 224, 224),
        validation: bool = False,
    ) -> DatasetDefault:
        """
        Creates Fuse Dataset single object (either for training, validation and test or user defined set)

        :param data_dir:                    dataset root path
        :param series_config                configuration of the selected series from the ukbb zip
        :param input_source_gt              dataframe with ground truth file
        :param cache_dir:                   Optional, name of the cache folder
        :param reset_cache:                 Optional,specifies if we want to clear the cache first
        :param num_workers: number of processes used for caching
        :param sample_ids: dataset including the specified sample_ids or None for all the samples. sample_id is case_{id:05d} (for example case_00001 or case_00100).
        :param train: True if used for training  - adds augmentation operations to the pipeline
        :return: DatasetDefault object
        """
        if isinstance(csv_path, str):
            df = pd.read_csv(csv_path)
        else:
            df = csv_path

        if sample_ids is None:
            sample_ids = SegOAI.sample_ids(df)

        static_pipeline = SegOAI.static_pipeline(df)
        dynamic_pipeline = SegOAI.dynamic_pipeline(
            validation=validation, resize_to=resize_to
        )

        if cache_dir is not None:
            cacher = SamplesCacher(
                "oai",
                static_pipeline,
                cache_dirs=[cache_dir],
                restart_cache=reset_cache,
                audit_first_sample=False,
                audit_rate=None,
                workers=num_workers,
                use_pipeline_hash=False,
            )

            my_dataset = DatasetDefault(
                sample_ids=sample_ids,
                static_pipeline=static_pipeline,
                dynamic_pipeline=dynamic_pipeline,
                cacher=cacher,
            )
        else:
            my_dataset = DatasetDefault(
                sample_ids=sample_ids,
                static_pipeline=static_pipeline,
                dynamic_pipeline=dynamic_pipeline,
                cacher=None,
                allow_uncached_sample_morphing=False,
            )

        my_dataset.create(num_workers=num_workers)
        return my_dataset



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""
from __future__ import annotations
from _collections_abc import dict_items, dict_keys
import difflib

import copy
import types
from numpy import ndarray
from torch import Tensor
from typing import (
    Any,
    Callable,
    Iterator,
    Optional,
    Sequence,
    Union,
    List,
    MutableMapping,
)


class NDict(dict):
    """
    (Lazy) N(ested)Dict - wraps a python dict, and allows to access "nested" elements via '.' separated key desc

    NOTE: assumes that all keys (including nested) are:
        1. strings
        2. do not contain '.' within a single key, as '.' is used as a special symbol for accessing deeper level nested dict.

    For example:

        nx = NDict(x)
        nx['a.b'] = 14
        assert nx['a.b'] == 14

        if the result is a non-leaf, you will get a NDict instance, for example
        nx = NDict()
        nx["a.d.zz] = 42
        assert nx['a']['d.zz'] == 42

    """

    def __init__(
        self,
        dict_like: Union[dict, tuple, types.GeneratorType, NDict, None] = None,
        already_flat: bool = False,
    ):
        """
        :param dict_like: the data with which to populate the nested dictionary, in case of NDict it acts as view constructor,
            otherwise we just set all the keys and values using the setitem function
        :param already_flat: optimization option. set to True only if you are sure that the input "dict_like" is a dict without nested dictionaries.
        """

        if dict_like is None:
            self._stored = dict()

        elif isinstance(dict_like, NDict):
            self._stored = dict_like._stored

        elif isinstance(dict_like, dict) and already_flat:
            self._stored = dict_like

        elif dict_like is not None:
            self._stored = dict()
            if not isinstance(dict_like, MutableMapping):
                dict_like = dict(dict_like)
            for k, d in dict_like.items():
                self[k] = d

        else:
            raise Exception(f"Not supported dict_like type: {type(dict_like)}")

    def to_dict(self) -> dict:
        """
        converts to standard python dict
        """
        return self._stored

    def clone(self, deepcopy: bool = True) -> NDict:
        """
        does a deep or a shallow copy, shallow copy means only top level keys are copied and the values are only referenced.
        in deep copy, all values are copied recursively

        :param deepcopy: if true, does deep copy, otherwise does a shallow copy
        """
        if not deepcopy:
            return NDict(copy.copy(self._stored), already_flat=True)
        else:
            return NDict(copy.deepcopy(self._stored), already_flat=True)

    def flatten(self) -> NDict:
        """
        Legacy from previous implementation of NDict where we store the data with nested dictionaries.
        currently we store the data with one flat dictionary
        """
        return self

    def keypaths(self) -> List[str]:
        """
        :return: a list of keypaths (i.e. "a.b.c.d") to all values in the nested dict
        """
        return list(self._stored.keys())

    def keys(self) -> dict_keys:
        """
        returns keypaths (threat as a "flat" dictionary)
        """
        return self._stored.keys()

    def top_level_keys(self) -> List[str]:
        """
        return top-level keys.

        Example:
            >>> ndict = NDict()
            >>> ndict["a1.b.c"] = 42
            >>> ndict["a2.b.h"] = 23
            >>> ndict.top_level_keys()
            ['a1', 'a2']
        """
        top_level_keys = {key.split(".")[0] for key in self.keys()}
        return list(top_level_keys)

    def values(self) -> dict_items:
        return self._stored.values()

    def items(self) -> dict_items:
        return self._stored.items()

    def merge(self, other: MutableMapping) -> None:
        """
        inplace merge between self and other.
        """
        for k, v in other.items():
            self[k] = v
        return

    def __getitem__(self, key: str) -> Any:
        """
        "traverses" the nested dict by the path extracted from splitting the key on '.'.
        if key not found, shows the possible closest options.
        if a prefix is given (the key exists but it is not a leaf), returns the corresponding sub dictionary

        :param key: dot delimited keypath into the nested dict
        """

        try:
            # the key is a full key for a value
            return self._stored[key]

        except KeyError:
            # the key is a prefix for other value(s)
            sub_dict = self.get_sub_dict(key)

            if sub_dict is not None:
                return sub_dict
            else:
                raise NestedKeyError(key, self)

    def get_sub_dict(self, key: str) -> Union[NDict, None]:
        """
        returns a copy of the "sub-dict" give a sub-key.
        if a sub-dict wasn't found, returns None

        Example:
            >>> ndict = NDict()
            >>> ndict["a.b.c"] = "x"
            >>> ndict["a.b.d"] = "y"
            >>> ndict["a.b.e"] = "z"
            >>> ndict.get_sub_dict("a.b")
            {'c': 'x', 'd': 'y', 'e': 'z'}
        """
        res = NDict()
        prefix_key = key + "."
        suffix_key = None
        for kk in self.keys():
            if kk.startswith(prefix_key):
                suffix_key = kk[len(prefix_key) :]
                res[suffix_key] = self[kk]

        if suffix_key is None:
            return None

        return res

    def __setitem__(self, key: str, value: Any) -> None:
        """
        sets item in the dictionary.
        if the item is a dict like object, it will parse the values using the delimiter "."

        Example:
            ndict = NDict()
            ndict["a"] = {"b" : {"c" : 42}}

            assert ndict["a.b.c"] == 42

        :param key: the keypath
        :param value: value to set
        """
        # if value is dictionary add to self key by key to avoid from keys with delimiter "."
        if isinstance(value, MutableMapping):
            for sub_key in value:
                self[f"{key}.{sub_key}"] = value[sub_key]
            return
        self._stored[key] = value

    def __delitem__(self, key: str) -> None:
        """
        deletes self[key] item(s).
        if key is a prefix to a sub-dict, deletes the entire sub-dict.
        if key is a key for a value AND a sub-dict, deletes BOTH
        """
        deleted = False

        # delete specific (key, value)
        if key in self._stored:
            del self._stored[key]
            deleted = True

        # delete entire branch
        for kk in list(self.keys()):
            if kk.startswith(f"{key}."):
                del self[kk]
                deleted = True

        if not deleted:
            raise NestedKeyError(key, self)

    def get_closest_keys(self, key: str, n: int = 1) -> List[str]:
        """
        For a given keypath, returns the closest key(s) in the current nested dict.
        if key exists in the dictionary
        :param key: string as a key
        :param n: the amount closest key(s) to return
        """
        if key in self._stored:
            return [key]

        total_sub_keys = []

        for kk in self.keys():
            splitted_key = kk.split(".")
            for i in range(len(splitted_key)):
                total_sub_keys.append(".".join(splitted_key[: i + 1]))

        closest_keys = difflib.get_close_matches(key, total_sub_keys, n=n, cutoff=0)
        return closest_keys

    def pop(self, key: str) -> Any:
        """
        returns the value self[key] and removes the key from the dict.

        :param key: the key to return and remove
        """
        res = self[key]
        del self[key]
        return res

    def indices(self, indices: ndarray) -> dict:
        """
        Extract the specified indices from each element in the dictionary (if possible)

        :param nested_dict: input dict
        :param indices: indices to extract. Either list of indices or boolean numpy array
        :return: NDict with only the required indices
        """
        new_dict = {}
        all_keys = self.keys()
        for key in all_keys:
            try:
                value = self[key]
                if isinstance(value, (ndarray, Tensor)):
                    new_value = value[indices]
                elif isinstance(value, Sequence):
                    new_value = [item for i, item in enumerate(value) if indices[i]]
                else:
                    new_value = value
                new_dict[key] = new_value
            except:
                print(f"failed to process key {key}")
                raise
        return new_dict

    def apply_on_all(self, apply_func: Callable, *args: Any) -> None:
        """
        Inplace apply specified function on all the dictionary's values

        :param apply_func: function to apply
        :param args: custom arguments for the 'apply_func' function
        """
        all_keys = self.keys()
        for key in all_keys:
            new_value = apply_func(self[key], *args)
            self[key] = new_value

    def __reduce__(self) -> Union[str, tuple]:
        return super().__reduce__()

    def __iter__(self) -> Iterator:
        return iter(self._stored)

    def __len__(self) -> int:
        return len(self._stored)

    def __str__(self) -> str:
        return str(self._stored)

    def __repr__(self) -> str:
        return repr(self._stored)

    def __contains__(self, key: str) -> bool:
        if key in self._stored:
            return True

        key_dot = key + "."
        # iterate over all keys and looking for a match
        for kk in self.keys():
            if kk.startswith(key_dot):
                return True

        # a match wasn't found
        return False

    def get(self, key: str, default_value: Any = None) -> Any:
        if key not in self:
            return default_value
        return self[key]

    def get_multi(self, keys: Optional[List[str]] = None) -> NDict:
        """
        returns a subset of the dict with the specified keys

        :param keys: keys to keep in the returned ndict
        """
        if keys is None:
            keys = self.keys()  # take all keys

        ans = NDict()
        for k in keys:
            curr = self[k]
            ans[k] = curr

        return ans

    def unflatten(self) -> dict:
        """
        returns a nested dictionaries

        Example:

        >>> ndict = NDict({"a.b.c":42, "a.b.d": 23})
        >>> ndict.unflatten()
        {'a': {'b': {'d': 23, 'c': 42}}}
        """

        return NDict._unflatten_static(data_dict=self)

    @staticmethod
    def _unflatten_static(data_dict: NDict) -> dict:
        res = dict()
        for top_key in data_dict.top_level_keys():
            value = data_dict[top_key]
            if isinstance(value, NDict):
                res[top_key] = NDict._unflatten_static(value)
            else:
                res[top_key] = value

        return res

    def update(self, dict_like: MutableMapping) -> None:
        self.merge(dict_like)

    def print_tree(self, print_values: bool = False) -> None:
        """
        print the inner structure of the nested dict with a tree-like structure.

        :param print_values: set to True in order to also print ndict's stored values


        Example:

            >>> ndict = NDict()
            >>> ndict["data.input.drug"] = "this_is_a_drug_seq"
            >>> ndict["data.input.target"] = "this_is_a_target_seq"
            >>>
            >>> ndict.print_tree()
            --- data
            ------ input
            --------- drug
            --------- target
            >>>
            >>> ndict.print_tree(print_values=True)
            --- data
            ------ input
            --------- drug -> this_is_a_drug_seq
            --------- target -> this_is_a_target_seq

        """
        print(self.get_tree(print_values=print_values))

    def get_tree(self, print_values: bool = False) -> str:
        """
        returns a string of the NDict object in a tree-like structure. See 'print_tree'.
        """
        unflatten_dict = self.unflatten()
        tree = self._get_tree_str_static(unflatten_dict, print_values=print_values)
        return tree[:-1] if len(tree) > 0 else tree

    @staticmethod
    def _get_tree_str_static(
        data_dict: dict, level: int = 0, print_values: bool = False
    ) -> None:
        """
        static-method to print the inner structure of a dict in a tree-like structure.

        :param level: current recursive level inside the ndict
        :param print_values: set to True in order to also print ndict's stored values
        """
        res = ""
        keys = data_dict.keys()
        level += 1
        for key in keys:
            if isinstance(data_dict[key], dict):
                res += " ".join(["---" * level, key + "\n"])
                res += NDict._get_tree_str_static(
                    data_dict[key], level, print_values=print_values
                )
            else:
                if print_values:
                    res += " ".join(
                        ["---" * level, key, "->", str(data_dict[key]) + "\n"]
                    )
                else:
                    res += " ".join(["---" * level, key + "\n"])

        return res

    def describe(self) -> None:
        for k in self.keys():
            print(f"{k}")
            val = self[k]
            print(f"\ttype={type(val)}")
            if hasattr(val, "shape"):
                print(f"\tshape={val.shape}")


class NestedKeyError(KeyError):
    def __init__(self, key: str, d: NDict) -> None:
        closest_keys = d.get_closest_keys(key, n=3)
        if len(closest_keys) == 0:
            error_str = f"Error: key {key} does not exist\n. All keys: {d.keys()}"
        else:
            error_str = f"Error: key {key} does not exist\n. Top {len(closest_keys)} Closest key are: {closest_keys}. All keys: {d.keys()}"
        print(error_str)
        super().__init__(error_str)



from typing import Dict, Callable
from fuse.utils import gpu as FuseUtilsGPU
from sklearn.model_selection import KFold
from functools import partial
from multiprocessing import Process, Queue
from typing import Sequence, Union
import os
from fuse.eval.metrics.classification.metrics_ensembling_common import MetricEnsemble
from collections import OrderedDict
from fuse.eval.evaluator import EvaluatorDefault
from fuse.utils.file_io.file_io import create_or_reset_dir
from fuse.utils.rand.seed import Seed
from multiprocessing import set_start_method

set_start_method("spawn", force=True)


def ensemble(
    test_dirs: Sequence[str],
    test_infer_filename: str,
    pred_key: str,
    target_key: str,
    ensembled_output_file: str,
) -> None:
    ensembled_output_dir = os.path.dirname(ensembled_output_file)
    create_or_reset_dir(ensembled_output_dir, force_reset=True)
    test_infer_filenames = [os.path.join(d, test_infer_filename) for d in test_dirs]
    # define data for ensemble metric
    data = {str(k): test_infer_filenames[k] for k in range(len(test_infer_filenames))}

    # list of metrics
    metrics = OrderedDict(
        [
            (
                "ensemble",
                MetricEnsemble(
                    pred_keys=[str(i) + "." + pred_key for i in range(len(test_dirs))],
                    target="0." + target_key,
                    output_file=ensembled_output_file,
                    output_pred_key=pred_key,
                    output_target_key=target_key,
                ),
            ),
        ]
    )

    evaluator = EvaluatorDefault()
    _ = evaluator.eval(ids=None, data=data, metrics=metrics)


def runner_wrapper(
    q_resources: Queue,
    rep_index: int,
    deterministic_mode: bool,
    fs: Union[Sequence[Callable], Callable],
    *f_args: tuple,
    **f_kwargs: dict,
) -> None:
    resource = q_resources.get()
    print(f"Using GPUs: {resource}")
    FuseUtilsGPU.choose_and_enable_multiple_gpus(
        len(resource), force_gpus=list(resource)
    )
    _ = Seed.set_seed(rep_index, deterministic_mode=deterministic_mode)
    if isinstance(fs, Sequence):
        for f, prev_arg, last_arg in zip(fs, f_args[-2], f_args[-1]):
            f(*(f_args[:-2] + (prev_arg,) + (last_arg,)), **f_kwargs)
    else:
        f(*f_args, **f_kwargs)
    print(f"Done with GPUs: {resource} - adding them back to the queue")
    q_resources.put(resource)


def train_wrapper(
    sample_ids_per_fold: Sequence,
    cv_index: int,
    rep_index: int,
    dataset_func: Callable,
    dataset_params: dict,
    paths: dict,
    func: Callable,
    params: dict,
) -> None:
    paths_train = paths.copy()

    # set parameters specific to this fold:
    paths_train["model_dir"] = os.path.join(
        paths["model_dir"], "rep_" + str(rep_index), str(cv_index)
    )
    paths_train["cache_dir"] = os.path.join(
        paths["cache_dir"], "rep_" + str(rep_index), str(cv_index)
    )

    # generate data for this fold:
    train_dataset, validation_dataset = dataset_func(
        train_val_sample_ids=sample_ids_per_fold,
        paths=paths_train,
        params=dataset_params,
    )

    # call project specific train_func:
    func(
        train_dataset=train_dataset,
        validation_dataset=validation_dataset,
        paths=paths_train,
        train_params=params,
    )


def infer_wrapper(
    sample_ids_per_fold: Sequence,
    cv_index: int,
    rep_index: int,
    dataset_func: Callable,
    dataset_params: dict,
    paths: dict,
    func: Callable,
    params: dict,
) -> None:
    paths_infer = paths.copy()

    # set parameters specific to this fold, and generate data:
    paths_infer["model_dir"] = os.path.join(
        paths["model_dir"], "rep_" + str(rep_index), str(cv_index)
    )
    if sample_ids_per_fold is None:  # test mode
        paths_infer["inference_dir"] = os.path.join(
            paths["inference_dir"], "test", "rep_" + str(rep_index), str(cv_index)
        )
        _, dataset = dataset_func(
            train_val_sample_ids=None, paths=paths_infer, params=dataset_params
        )
    else:
        paths_infer["inference_dir"] = os.path.join(
            paths["inference_dir"], "validation", "rep_" + str(rep_index), str(cv_index)
        )
        _, dataset = dataset_func(
            train_val_sample_ids=sample_ids_per_fold,
            paths=paths_infer,
            params=dataset_params,
        )

    # call project specific infer_func:
    func(dataset=dataset, paths=paths_infer, infer_params=params)


def eval_wrapper(
    sample_ids_per_fold: Sequence,
    cv_index: int,
    rep_index: int,
    dataset_func: Callable,
    dataset_params: dict,
    paths: dict,
    func: Callable,
    params: dict,
) -> None:
    paths_eval = paths.copy()

    if sample_ids_per_fold is None:  # test mode
        paths_eval["inference_dir"] = os.path.join(
            paths["inference_dir"], "test", "rep_" + str(rep_index), str(cv_index)
        )
        paths_eval["eval_dir"] = os.path.join(
            paths["eval_dir"], "test", "rep_" + str(rep_index), str(cv_index)
        )
    else:
        paths_eval["inference_dir"] = os.path.join(
            paths["inference_dir"], "validation", "rep_" + str(rep_index), str(cv_index)
        )
        paths_eval["eval_dir"] = os.path.join(
            paths["eval_dir"], "validation", "rep_" + str(rep_index), str(cv_index)
        )

    # call project specific eval_func:
    func(paths=paths_eval, eval_params=params)


def run(
    num_folds: int,
    num_folds_used: int,
    num_gpus_total: int,
    num_gpus_per_split: int,
    num_repetitions: int,
    dataset_func: Callable,
    train_func: Callable,
    infer_func: Callable,
    eval_func: Callable,
    dataset_params: Dict = None,
    train_params: Dict = None,
    infer_params: Dict = None,
    eval_params: Dict = None,
    paths: Dict = None,
    deterministic_mode: bool = True,
    sample_ids_per_fold: Sequence = None,
) -> None:
    """
    ML pipeline - run a full ML experiment pipeline which consists of training on multiple
    cross validation folds, validation inference and evaluation, inference and evaluation
    of each fold's model on a held out test set, ensembling the models trained on different folds,
    and evaluation of the ensembled model.
    The whole process can also be repeated multiple times, to evaluate variability due to different
    random seed.
    Multiple cross validation folds run simultaneously on available GPU resources.

    :param num_folds: Number of cross validation splits/folds.
    :param num_folds_used: Number of folds/splits to use.
        For example, for training a single model with 80% of the samples used
        for training and 20% for validation, set `num_folds=5` and `num_folds_used=1`.
        For running a 5-fold cross-validation scenario, set `num_folds=5` and `num_folds_used=5`.
    :param num_gpus_total: Number of GPUs to use in total for executing the pipeline.
    :param num_gpus_per_split: Number of GPUs to use for a single model training/inference.
    :param num_repetitions: Number of repetitions of the procedure with different random
        seeds. Note that this does not change the random decision on cross validation
        fold sample ids.
    :param dataset_func: Callable to a custom function that implements a dataset
        creation. Its input is a path to cache directory and it should return a
        train and test dataset.
    :param train_func: Callable to a custom function that executes model training.
    :param infer_func: Callable to a custom inference function.
    :param eval_func: Callable to a custom evaluation function.
    :param dataset_params: Dictionary that can contain any number of additional
        custom parameters for dataset_func.
    :param train_params: Dictionary that can contain any number of additional
        custom parameters for train_func.
    :param infer_params: Dictionary that can contain any number of additional
        custom parameters for infer_func.
    :param eval_params: Dictionary that can contain any number of additional
        custom parameters for eval_func.
    :param deterministic_mode: Whether to use PyTotch deterministic mode.
    :param sample_ids_per_fold: May contain a sequence of array pairs denoting sample ids
        for pre-defined train/validation splits. If this parameter is set to `None`,
        the splits are decided at random.
    """
    available_gpu_ids = FuseUtilsGPU.get_available_gpu_ids()
    if num_gpus_total < len(available_gpu_ids):
        available_gpu_ids = available_gpu_ids[0:num_gpus_total]
    # group gpus into chunks of size num_gpus_per_split
    gpu_resources = [
        available_gpu_ids[i : i + num_gpus_per_split]
        for i in range(0, len(available_gpu_ids), num_gpus_per_split)
    ]

    # create a queue of gpu chunks (resources)
    q_resources = Queue()
    for r in gpu_resources:
        q_resources.put(r)

    if sample_ids_per_fold is None:
        dataset, test_dataset = dataset_func(paths=paths, params=dataset_params)
        # the split decision should be the same regardless of repetition index
        kfold = KFold(n_splits=num_folds, shuffle=True, random_state=1234)
        sample_ids_per_fold = [item for item in kfold.split(dataset)]
    else:
        assert num_folds == len(sample_ids_per_fold)

    for rep_index in range(num_repetitions):
        # run training, inference and evaluation on all cross validation folds in parallel
        # using the available gpu resources:
        runner = partial(
            runner_wrapper,
            q_resources,
            rep_index,
            deterministic_mode,
            [train_wrapper, infer_wrapper, eval_wrapper],
        )

        # create process per fold
        processes = [
            Process(
                target=runner,
                args=(
                    ids,
                    cv_index,
                    rep_index,
                    dataset_func,
                    dataset_params,
                    paths,
                    [train_func, infer_func, eval_func],
                    [train_params, infer_params, eval_params],
                ),
            )
            for (ids, cv_index) in zip(sample_ids_per_fold, range(num_folds))
        ][0:num_folds_used]
        for p in processes:
            p.start()

        for p in processes:
            p.join()
            p.close()

        # infer and eval each split's model on test set:
        runner = partial(
            runner_wrapper,
            q_resources,
            rep_index,
            deterministic_mode,
            [infer_wrapper, eval_wrapper],
        )
        # create process per fold
        processes = [
            Process(
                target=runner,
                args=(
                    None,
                    cv_index,
                    rep_index,
                    dataset_func,
                    dataset_params,
                    paths,
                    [infer_func, eval_func],
                    [infer_params, eval_params],
                ),
            )
            for cv_index in range(num_folds)
        ][0:num_folds_used]
        for p in processes:
            p.start()

        for p in processes:
            p.join()
            p.close()

        # generate ensembled predictions:
        test_dirs = [
            os.path.join(
                paths["inference_dir"], "test", "rep_" + str(rep_index), str(cv_index)
            )
            for cv_index in range(num_folds)
        ][0:num_folds_used]
        test_infer_filename = "infer.gz"
        ensembled_output_file = os.path.join(
            paths["inference_dir"],
            "test",
            "rep_" + str(rep_index),
            "ensemble",
            "infer.gz",
        )

        ensemble(
            test_dirs,
            test_infer_filename,
            infer_params["pred_key"],
            dataset_params["target_key"],
            ensembled_output_file,
        )

        # evaluate ensemble:
        paths_eval = paths.copy()
        paths_eval["inference_dir"] = os.path.join(
            paths["inference_dir"], "test", "rep_" + str(rep_index), "ensemble"
        )
        paths_eval["eval_dir"] = os.path.join(
            paths["eval_dir"], "test", "rep_" + str(rep_index), "ensemble"
        )

        _ = Seed.set_seed(rep_index, deterministic_mode=deterministic_mode)
        eval_func(paths=paths_eval, eval_params=eval_params)



# type: ignore
"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on Jan 09, 2023

"""

import pickle
import tqdm
from typing import Optional, Sequence, List, Union, Tuple
from collections import Counter, defaultdict

# dictionary of special tokens that will be used in a generation of
# patients trajectory sequence
special_tokens = {
    "padding": "PAD",
    "unknown": "UNK",
    "separator": "SEP",
    "cls": "CLS",
    "separator_static": "SEP_STATIC",
}


def seq_translate(tokens: List[str], translate_dict: dict) -> Tuple[List[str]]:
    """
    returns a list of tokens translated using translate_dict
    :param tokens:
    :param translate_dict:
    :return:
    """
    return (
        [
            translate_dict.get(token, translate_dict[special_tokens["unknown"]])
            for token in tokens
        ],
    )


def position_idx(
    tokens: Sequence[str], symbol: str = special_tokens["separator"]
) -> List[int]:
    """
    Given a sequence of codes divided into groups (visits)
     by symbol ('SEP') tokens, returns a sequence of the same
     size of visit indices.
    :param tokens:
    :param symbol:
    :return:
    """
    group_inds = []
    flag = 0
    for token in tokens:
        group_inds.append(flag)
        if token == symbol:
            flag += 1
    return group_inds


def seq_pad(
    tokens: Sequence[str], max_len: int, symbol: str = special_tokens["padding"]
) -> List[str]:
    """
    Returns a list of tokens padded by symbol to length max_len.
    :param tokens:
    :param max_len:
    :param symbol:
    :return:
    """
    token_len = len(tokens)
    if token_len < max_len:
        return list(tokens) + [symbol] * (max_len - token_len)

    else:
        return tokens[:max_len]


class TorchVocab(object):
    """Defines a vocabulary object that will be used to numericalize a field.
    Attributes:
        freqs: A collections.Counter object holding the frequencies of tokens
            in the data used to build the Vocab.
        stoi: A collections.defaultdict instance mapping token strings to
            numerical identifiers.
        itos: A list of token strings indexed by their numerical identifiers.
    """

    def __init__(
        self,
        counter: Counter,
        max_size: Optional[int] = None,
        min_freq: int = 1,
        specials: Sequence[str] = ["<pad>", "<oov>"],
        vectors=None,
        unk_init=None,
        vectors_cache=None,
    ) -> None:
        """Create a Vocab object from a collections.Counter.
        Arguments:
            counter: collections.Counter object holding the frequencies of
                each value found in the data.
            max_size: The maximum size of the vocabulary, or None for no
                maximum. Default: None.
            min_freq: The minimum frequency needed to include a token in the
                vocabulary. Values less than 1 will be set to 1. Default: 1.
            specials: The list of special tokens (e.g., padding or eos) that
                will be prepended to the vocabulary in addition to an <unk>
                token. Default: ['<pad>']
            vectors: One of either the available pretrained vectors
                or custom pretrained vectors (see Vocab.load_vectors);
                or a list of aforementioned vectors
            unk_init (callback): by default, initialize out-of-vocabulary
                word vectors to zero vectors; can be any function that takes
                in a Tensor and returns a Tensor of the same size.
                Default: torch.Tensor.zero_vectors_cache: directory for cached
                vectors. Default: '.vector_cache'
        """
        self.freqs = counter
        counter = counter.copy()
        min_freq = max(min_freq, 1)

        self.itos = list(specials)
        # frequencies of special tokens are not counted when
        # building vocabulary in frequency order
        for tok in specials:
            del counter[tok]

        max_size = None if max_size is None else max_size + len(self.itos)

        # sort by frequency, then alphabetically
        words_and_frequencies = sorted(counter.items(), key=lambda tup: tup[0])
        words_and_frequencies.sort(key=lambda tup: tup[1], reverse=True)

        for word, freq in words_and_frequencies:
            if freq < min_freq or len(self.itos) == max_size:
                break
            self.itos.append(word)

        # stoi is simply a reverse dict for itos
        self.stoi = {tok: i for i, tok in enumerate(self.itos)}

        self.vectors = None
        if vectors is not None:
            self.load_vectors(vectors, unk_init=unk_init, cache=vectors_cache)
        else:
            assert unk_init is None and vectors_cache is None

    def __eq__(self, other: "TorchVocab") -> bool:
        if self.freqs != other.freqs:
            return False
        if self.stoi != other.stoi:
            return False
        if self.itos != other.itos:
            return False
        if self.vectors != other.vectors:
            return False
        return True

    def __len__(self) -> int:
        return len(self.itos)

    def vocab_rerank(self) -> None:
        self.stoi = {word: i for i, word in enumerate(self.itos)}

    def extend(self, v: "TorchVocab", sort: bool = False) -> None:
        words = sorted(v.itos) if sort else v.itos
        for w in words:
            if w not in self.stoi:
                self.itos.append(w)
                self.stoi[w] = len(self.itos) - 1


class Vocab(TorchVocab):
    def __init__(
        self, counter: Counter, max_size: Optional[int] = None, min_freq: int = 1
    ):
        self.pad_index = 0
        self.unk_index = 1
        self.eos_index = 2
        self.sos_index = 3
        self.mask_index = 4
        super().__init__(
            counter,
            specials=list(special_tokens.values()),
            max_size=max_size,
            min_freq=min_freq,
        )

    def to_seq(self, sentece, seq_len, with_eos=False, with_sos=False) -> list:
        pass

    def from_seq(self, seq, join=False, with_pad=False):
        pass

    @staticmethod
    def load_vocab(vocab_path: str) -> "Vocab":
        with open(vocab_path, "rb") as f:
            return pickle.load(f)

    def save_vocab(self, vocab_path: str) -> None:
        with open(vocab_path, "wb") as f:
            pickle.dump(self, f)


# Building Vocab with text files
class WordVocab(Vocab):
    def __init__(
        self,
        texts: List[Union[List[str], str]],
        max_size: Optional[int] = None,
        min_freq: int = 1,
    ):
        print("Building Vocab")
        counter = Counter()
        for line in tqdm.tqdm(texts):
            if isinstance(line, list):
                words = line
            else:
                words = line.replace("\n", "").replace("\t", "").split()

            for word in words:
                counter[word] += 1
        super().__init__(counter, max_size=max_size, min_freq=min_freq)

    def to_seq(
        self,
        sentence: Union[str, List[str]],
        seq_len: Optional[int] = None,
        with_eos: bool = False,
        with_sos: bool = False,
        with_len: bool = False,
    ) -> Union[str, Tuple[str, int]]:
        if isinstance(sentence, str):
            sentence = sentence.split()

        seq = [self.stoi.get(word, self.unk_index) for word in sentence]

        if with_eos:
            seq += [self.eos_index]  # this would be index 1
        if with_sos:
            seq = [self.sos_index] + seq

        origin_seq_len = len(seq)

        if seq_len is None:
            pass
        elif len(seq) <= seq_len:
            seq += [self.pad_index for _ in range(seq_len - len(seq))]
        else:
            seq = seq[:seq_len]

        return (seq, origin_seq_len) if with_len else seq

    def from_seq(
        self, seq: Sequence, join: bool = False, with_pad: bool = False
    ) -> str:
        words = [
            self.itos[idx] if idx < len(self.itos) else "<%d>" % idx
            for idx in seq
            if not with_pad or idx != self.pad_index
        ]

        return " ".join(words) if join else words

    def get_stoi(self) -> defaultdict:
        return self.stoi

    def get_itos(self) -> List[str]:
        return self.itos

    @staticmethod
    def load_vocab(vocab_path: str) -> "WordVocab":
        with open(vocab_path, "rb") as f:
            return pickle.load(f)



import hydra
from omegaconf import DictConfig, OmegaConf
from fuse.data.tokenizers.modular_tokenizer.modular_tokenizer import ModularTokenizer
from typing import Dict, Any


@hydra.main(
    config_path="../pretrained_tokenizers/configs",
    config_name="tokenizer_config",
    version_base=None,
)
def main(cfg: DictConfig) -> None:
    print(str(cfg))

    cfg = hydra.utils.instantiate(cfg)
    tmp = OmegaConf.to_object(cfg)
    cfg_raw: Dict[str, Any] = tmp

    cfg_tokenizer: Dict[str, Any] = cfg_raw["data"]["tokenizer"]
    t_mult = ModularTokenizer(
        **cfg_tokenizer,
    )

    t_mult.save(path=cfg_raw["data"]["tokenizer"]["out_path"])


if __name__ == "__main__":
    main()



"""
(C) Copyright 2021 IBM Corp.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at
   http://www.apache.org/licenses/LICENSE-2.0
Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
Created on June 30, 2021
"""

import os
import copy
from pathlib import Path
import pickle
import pandas as pd
import numpy as np
from fuse.utils.utils_debug import FuseDebug
from fuse.utils.gpu import choose_and_enable_multiple_gpus

import logging
import torch.optim as optim
from torch.utils.data.dataloader import DataLoader

from fuse.utils.utils_logger import fuse_logger_start
from fuse.utils import NDict
from fuse.data.utils.samplers import BatchSamplerDefault
from fuse.data.utils.collates import CollateDefault
from fuse.data.utils.split import dataset_balanced_division_to_folds
from fuse.dl.losses.loss_default import LossDefault

# from report_guided_annotation import extract_lesion_candidates
from fuse_examples.imaging.segmentation.picai.unet import UNet
from fuse.eval.metrics.segmentation.metrics_segmentation_common import MetricDice
from collections import OrderedDict
from fuse.dl.losses.segmentation.loss_dice import BinaryDiceLoss

# from fuse.eval.metrics.detection.metrics_detection_common import MetricDetectionPICAI
from fuseimg.datasets.picai import PICAI
from fuse.dl.lightning.pl_funcs import convert_predictions_to_dataframe
from pytorch_lightning import Trainer
from fuse.dl.lightning.pl_module import LightningModuleDefault
from fuse.utils.file_io.file_io import create_dir, load_pickle, save_dataframe
from fuse.eval.evaluator import EvaluatorDefault

# from picai_baseline.unet.training_setup.default_hyperparam import \
#     get_default_hyperparams
# from picai_baseline.unet.training_setup.neural_network_selector import \
#     neural_network_for_run
import torch
import hydra
from omegaconf import DictConfig, OmegaConf

# from monai.losses import DiceFocalLoss

# assert (
#     "PICAI_DATA_PATH" in os.environ
# ), "Expecting environment variable PICAI_DATA_PATH to be set. Follow the instruction in example README file to download and set the path to the data"
##########################################
# Debug modes
##########################################
mode = "default"  # Options: 'default', 'debug'. See details in FuseDebug
debug = FuseDebug(mode)


def create_model(unet_kwargs: dict) -> torch.nn.Module:
    """
    creates the model
    """
    model = UNet(
        input_name="data.input.img_t2w",
        seg_name="model.seg",
        unet_kwargs=unet_kwargs,
    )
    return model


#################################
# Train Template
#################################
def run_train(paths: NDict, train: NDict) -> None:
    # ==============================================================================
    # Logger
    # ==============================================================================
    # fuse_logger_start(output_path=paths["model_dir"], console_verbose_level=logging.INFO)
    lgr = logging.getLogger("Fuse")

    # Download data
    # ==============================================================================
    # Model
    # ==============================================================================
    lgr.info("Model:", {"attrs": "bold"})

    model = create_model(train["unet_kwargs"])
    lgr.info("Model: Done", {"attrs": "bold"})

    lgr.info("\nFuse Train", {"attrs": ["bold", "underline"]})

    lgr.info(f'model_dir={paths["model_dir"]}', {"color": "magenta"})
    lgr.info(f'cache_dir={paths["cache_dir"]}', {"color": "magenta"})

    #### Train Data
    # split to folds randomly - temp
    dataset_all = PICAI.dataset(
        paths=paths,
        cfg=train,
        reset_cache=False,
        train=True,
        run_sample=train["run_sample"],
    )
    folds = dataset_balanced_division_to_folds(
        dataset=dataset_all,
        output_split_filename=os.path.join(
            paths["data_misc_dir"], paths["data_split_filename"]
        ),
        id="data.patientID",
        keys_to_balance=["data.gt.classification"],
        nfolds=train["num_folds"],
        workers=train["num_workers"],
    )

    train_sample_ids = []
    for fold in train["train_folds"]:
        train_sample_ids += folds[fold]
    validation_sample_ids = []
    for fold in train["validation_folds"]:
        validation_sample_ids += folds[fold]

    train_dataset = PICAI.dataset(
        paths=paths,
        cfg=train,
        reset_cache=False,
        sample_ids=train_sample_ids,
        train=True,
    )

    validation_dataset = PICAI.dataset(
        paths=paths, cfg=train, reset_cache=False, sample_ids=validation_sample_ids
    )

    ## Create sampler
    lgr.info("- Create sampler:")
    sampler = BatchSamplerDefault(
        dataset=train_dataset,
        balanced_class_name="data.gt.classification",
        num_balanced_classes=2,
        batch_size=train["batch_size"],
        mode="approx",
        workers=train["num_workers"],
        balanced_class_weights=None,
    )

    lgr.info("- Create sampler: Done")

    ## Create dataloader
    train_dataloader = DataLoader(
        dataset=train_dataset,
        shuffle=False,
        drop_last=False,
        batch_sampler=sampler,
        collate_fn=CollateDefault(
            special_handlers_keys={
                "data.input.img": CollateDefault.pad_all_tensors_to_same_size
            },
            raise_error_key_missing=False,
        ),
        num_workers=train["num_workers"],
    )
    lgr.info("Train Data: Done", {"attrs": "bold"})

    #### Validation data
    lgr.info("Validation Data:", {"attrs": "bold"})

    ## Create dataloader
    validation_dataloader = DataLoader(
        dataset=validation_dataset,
        shuffle=False,
        drop_last=False,
        batch_sampler=None,
        batch_size=train["batch_size"],
        num_workers=train["num_workers"],
        collate_fn=CollateDefault(),
    )
    lgr.info("Validation Data: Done", {"attrs": "bold"})

    # ====================================================================================
    #  Loss and metrics
    # ====================================================================================
    # TODO - add a classification loss - add head to the bottom of the unet
    losses = {}
    losses["segmentation"] = LossDefault(
        pred="model.seg",
        target="data.gt.seg",
        callable=BinaryDiceLoss(),
    )
    train_metrics = {}

    validation_metrics = copy.deepcopy(
        train_metrics
    )  # use the same metrics in validation as well

    # either a dict with arguments to pass to ModelCheckpoint or list dicts for multiple ModelCheckpoint callbacks (to monitor and save checkpoints for more then one metric).
    best_epoch_source = dict(
        monitor="validation.losses.total_loss",
        mode="min",
    )

    # =====================================================================================
    #  Manager - Train
    #  Create a manager, training objects and run a training process.
    # =====================================================================================
    lgr.info("Train:", {"attrs": "bold"})

    # create optimizer
    optimizer = optim.Adam(
        model.parameters(),
        lr=train["learning_rate"],
        weight_decay=train["weight_decay"],
    )
    scheduler = optim.lr_scheduler.ReduceLROnPlateau(optimizer)

    lr_sch_config = dict(scheduler=scheduler, monitor="validation.losses.total_loss")

    # optimizier and lr sch - see pl.LightningModule.configure_optimizers return value for all options
    optimizers_and_lr_schs = dict(optimizer=optimizer, lr_scheduler=lr_sch_config)

    # create instance of PL module - FuseMedML generic version

    pl_module = LightningModuleDefault(
        model_dir=paths["model_dir"],
        model=model,
        losses=losses,
        train_metrics=train_metrics,
        validation_metrics=validation_metrics,
        best_epoch_source=best_epoch_source,
        optimizers_and_lr_schs=optimizers_and_lr_schs,
    )

    # create lightining trainer.
    pl_trainer = Trainer(
        default_root_dir=paths["model_dir"],
        max_epochs=train["trainer"]["num_epochs"],
        accelerator=train["trainer"]["accelerator"],
        devices=train["trainer"]["devices"],
        num_sanity_val_steps=-1,
    )

    # train from scratch
    pl_trainer.fit(
        pl_module,
        train_dataloader,
        validation_dataloader,
        ckpt_path=train["trainer"]["ckpt_path"],
    )
    lgr.info("Train: Done", {"attrs": "bold"})


######################################
# Inference Template
######################################
def run_infer(infer: NDict, paths: NDict, train: NDict) -> None:
    create_dir(paths["inference_dir"])
    #### Logger
    # fuse_logger_start(output_path=paths["inference_dir"], console_verbose_level=logging.INFO)
    lgr = logging.getLogger("Fuse")
    lgr.info("Fuse Inference", {"attrs": ["bold", "underline"]})
    infer_file = os.path.join(paths["inference_dir"], infer["infer_filename"])
    checkpoint_file = os.path.join(paths["model_dir"], infer["checkpoint"])
    lgr.info(f"infer_filename={checkpoint_file}", {"color": "magenta"})

    lgr.info("Model:", {"attrs": "bold"})

    model = create_model(train["unet_kwargs"])
    lgr.info("Model: Done", {"attrs": "bold"})
    ## Data
    folds = load_pickle(
        os.path.join(paths["data_misc_dir"], paths["data_split_filename"])
    )  # assume exists and created in train func

    infer_sample_ids = []
    for fold in infer["infer_folds"]:
        infer_sample_ids += folds[fold]

    test_dataset = PICAI.dataset(
        paths=paths,
        cfg=train,
        sample_ids=infer_sample_ids,
        reset_cache=False,
    )

    ## Create dataloader
    infer_dataloader = DataLoader(
        dataset=test_dataset,
        shuffle=False,
        drop_last=False,
        batch_sampler=None,
        batch_size=train["batch_size"],
        collate_fn=CollateDefault(),
        num_workers=infer["num_workers"],
    )
    # load python lightning module
    pl_module = LightningModuleDefault.load_from_checkpoint(
        checkpoint_file,
        model_dir=paths["model_dir"],
        model=model,
        map_location="cpu",
        strict=True,
    )
    # set the prediction keys to extract (the ones used be the evaluation function).
    pl_module.set_predictions_keys(
        ["model.seg", "data.gt.seg"]
    )  # which keys to extract and dump into file
    lgr.info("Test Data: Done", {"attrs": "bold"})
    # create lightining trainer.
    pl_trainer = Trainer(
        default_root_dir=paths["model_dir"],
        max_epochs=train["trainer"]["num_epochs"],
        accelerator=train["trainer"]["accelerator"],
        devices=train["trainer"]["devices"],
        num_sanity_val_steps=-1,
    )
    # create a trainer instance
    predictions = pl_trainer.predict(
        pl_module, infer_dataloader, return_predictions=True
    )

    with open(Path(infer_file).parents[0] / "infer.pickle", "wb") as f:
        pickle.dump(predictions, f)

    # convert list of batch outputs into a dataframe
    infer_df = convert_predictions_to_dataframe(predictions)
    save_dataframe(infer_df, infer_file)


######################################
# Analyze Template
######################################
def run_eval(paths: NDict, infer: NDict) -> None:
    fuse_logger_start(output_path=None, console_verbose_level=logging.INFO)
    lgr = logging.getLogger("Fuse")
    lgr.info("Fuse Eval", {"attrs": ["bold", "underline"]})

    # create evaluator
    evaluator = EvaluatorDefault()
    metrics = OrderedDict(
        [
            ("dice", MetricDice(pred="model.seg", target="data.gt.seg")),
        ]
    )
    # define iterator

    def data_iter() -> NDict:
        data_file = os.path.join(paths["inference_dir"], "infer.pickle")
        data = pd.read_pickle(data_file)
        for fold in data:
            for i, sample in enumerate(fold["id"]):
                sample_dict = {}
                sample_dict["id"] = sample
                sample_dict["model.seg"] = np.expand_dims(
                    fold["model.seg"][i][1], axis=0
                )
                sample_dict["model.seg"] = np.where(
                    sample_dict["model.seg"] > 0.7, 1, 0
                )
                sample_dict["data.gt.seg"] = fold["data.gt.seg"][i]
                yield sample_dict

    # run
    results = evaluator.eval(
        ids=None,
        data=data_iter(),
        batch_size=60,
        metrics=metrics,
        output_dir=paths["eval_dir"],
    )
    print(results)

    return results


@hydra.main(config_path="conf", config_name="config")
def main(cfg: DictConfig) -> None:
    cfg = NDict(OmegaConf.to_object(cfg))
    print(cfg)
    # uncomment if you want to use specific gpus instead of automatically looking for free ones
    force_gpus = None  # [0]
    choose_and_enable_multiple_gpus(cfg["train.trainer.devices"], force_gpus=force_gpus)

    # train
    if "train" in cfg["run.running_modes"]:
        run_train(NDict(cfg["paths"]), NDict(cfg["train"]))

    # infer
    if "infer" in cfg["run.running_modes"]:
        run_infer(NDict(cfg["infer"]), NDict(cfg["paths"]), NDict(cfg["train"]))

    # analyze
    if "eval" in cfg["run.running_modes"]:
        run_eval(NDict(cfg["paths"]), NDict(cfg["infer"]))


if __name__ == "__main__":
    main()



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import torch
import torch.nn as nn
from torch import Tensor
from typing import List


class BinaryDiceLoss(nn.Module):
    def __init__(self, eps: float = 1e-5, reduction: str = "mean"):
        """
        :param eps:         A float number to smooth loss, and avoid NaN error, default: 1
        :param reduction:   Reduction method to apply, return mean over batch if 'mean',
                            return sum if 'sum', return a tensor of shape [N,] if 'none'

        Returns:            Loss tensor according to arg reduction
        Raise:              Exception if unexpected reduction
        """
        super().__init__()
        self.reduction = reduction
        self.eps = eps

    def __call__(self, predict: Tensor, target: Tensor) -> Tensor:
        assert (
            predict.shape[0] == target.shape[0]
        ), "predict & target batch size don't match"
        reduce_axis: List[int] = torch.arange(2, len(predict.shape)).tolist()
        intersection = torch.sum(target * predict, dim=reduce_axis)
        ground_o = torch.sum(target, dim=reduce_axis)
        pred_o = torch.sum(predict, dim=reduce_axis)

        num = 2 * intersection + self.eps
        den = ground_o + pred_o + self.eps
        loss = 1 - num / den

        # return loss
        if self.reduction == "mean":
            return loss.mean()
        elif self.reduction == "sum":
            return loss.sum()
        elif self.reduction == "none":
            return loss
        else:
            raise Exception("Unexpected reduction {}".format(self.reduction))



from typing import Optional
import numpy as np
import torch
import torch.nn as nn

from fuse.dl.models.backbones.backbone_transformer import Transformer
from transformers.models.bert.modeling_bert import BertEncoder, BertPooler, BertConfig
from fuse.utils.ndict import NDict


class Embed(nn.Module):
    def __init__(
        self,
        n_vocab: int,
        emb_dim: int,
        key_in: str,
        key_out: str,
        **embedding_kwargs: dict
    ):
        super().__init__()
        self._emb_dim = emb_dim
        self._word_emb = nn.Embedding(n_vocab, self._emb_dim, **embedding_kwargs)
        self._key_in = key_in
        self._key_out = key_out

    def forward(self, batch_dict: NDict) -> NDict:
        tokens = batch_dict[self._key_in]
        tokens = tokens.to(device=next(iter(self._word_emb.parameters())).device)

        embds = self._word_emb(tokens)

        batch_dict[self._key_out] = embds

        return batch_dict


class WordDropout(nn.Module):
    def __init__(
        self,
        p_word_dropout: float,
        key_in: str,
        key_out: str,
        mask_value: int = 0,
        p_word_dropout_eval: Optional[float] = None,
    ):
        super().__init__()
        self._p = p_word_dropout
        self._p_eval = (
            p_word_dropout if p_word_dropout_eval is None else p_word_dropout_eval
        )
        self._key_in = key_in
        self._key_out = key_out
        self._mask_value = mask_value

    def forward(self, batch_dict: NDict) -> NDict:
        """
        Do word dropout: with prob `p_word_dropout`, set the word to '<unk>'.
        """
        x = batch_dict[self._key_in]

        data = x.clone().detach()

        # Sample masks: elems with val 1 will be set to <unk>
        p = self._p if self.training else self._p_eval

        mask = torch.from_numpy(
            np.random.choice(2, p=(1.0 - p, p), size=tuple(data.size())).astype("uint8")
        ).to(x.device)

        mask = mask.bool()
        # Set to <unk>
        data[mask] = self._mask_value

        batch_dict[self._key_out] = data
        return batch_dict


class TransformerEncoder(Transformer):
    def __init__(self, num_cls_tokens: int = 1, **kwargs: dict):
        super().__init__(num_cls_tokens=num_cls_tokens, **kwargs)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = super().forward(x)
        if self.num_cls_tokens == 1:
            return out[:, 0], out[:, 1:]
        else:
            return [out[:, i] for i in range(self.num_cls_tokens)] + [
                out[:, self.num_cls_tokens :]
            ]


class Bert(torch.nn.Module):
    def __init__(self, config: BertConfig):
        super().__init__()
        self.encoder = BertEncoder(config)
        self.pooler = BertPooler(config)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        encoded_layers = self.encoder(x)
        sequence_output = encoded_layers[0]  # this is the embedding of all tokens
        pooled_output = self.pooler(sequence_output)

        return pooled_output



import numpy as np
from typing import List, Tuple
import os


def num_available_cores(verbose: bool = True) -> int:
    if hasattr(os, "sched_getaffinity"):
        try:
            ans = len(os.sched_getaffinity(0))
            if verbose:
                print(
                    f"num_available_cores:: spotted affinity which restricts available cores. Returning {ans} cores"
                )
            return ans
        except Exception:
            pass

    ans = os.cpu_count()
    if verbose:
        print(f"num_available_cores:: Returning {ans} cores")
    return ans


def get_chunks_ranges(
    total: int, *, chunk_size: int = None, parts: int = None
) -> List[Tuple[int, int]]:
    """
    Creates "chunks" of work, useful when creating worker functions for run_multiprocessed
     where each workers gets a range ("chunk") to work on.

    Returns a list of tuples [start_index, end_index)

    Args:
        chunk_size - what is the size of a single chunks
        parts: how many chunks are desired

    for example: get_chunks_ranges(10, chunk_size = 3,) would return [(0,3), (3,6), (6,9), (9,10)]
        and get_chunks_ranges(10, parts=2) would return [(0,5),(5,10)]
    """

    assert (chunk_size is not None) ^ (
        parts is not None
    ), "Exactly one of chunk_size or parts must be provided"

    if chunk_size is not None:
        if chunk_size >= total:
            return [(0, total)]

        steps = np.arange(0, total, chunk_size, dtype=np.int64)
        ans = list(zip(steps[:-1], steps[1:]))
        if ans[-1][-1] < total:
            ans.append((ans[-1][-1], total))
        ans[-1] = (ans[-1][0], min(ans[-1][-1], total))
        return ans

    elif parts is not None:
        chunk_size = np.ceil(total / parts)
        return get_chunks_ranges(total, chunk_size=chunk_size)

    assert False, "should not reach here"


if __name__ == "__main__":
    ans = get_chunks_ranges(12, parts=2)
    z = 1



from fuse.utils.utils_logger import fuse_logger_start
import os
from glob import glob
import pandas as pd
from fuse.dl.models import ModelMultiHead
from fuse.utils import NDict

from fuse.dl.models.heads.head_dense_segmentation import HeadDenseSegmentation

from monai.losses import GeneralizedDiceLoss
import torch.nn as nn
import torch
from torch.utils.data.dataloader import DataLoader
from fuse.data.utils.collates import CollateDefault

from fuse.eval.metrics.segmentation.metrics_segmentation_common import MetricDice
import torch.optim as optim
from fuse.utils.rand.seed import Seed
import logging
from fuse.dl.losses.loss_default import LossDefault
from fuse.dl.lightning.pl_module import LightningModuleDefault
import pytorch_lightning as pl
from pytorch_lightning.callbacks import LearningRateMonitor
import hydra
from omegaconf import DictConfig
from clearml import Task
from fuse_examples.imaging.oai_example.data.seg_ds import SegOAI
from fuse.dl.models.backbones.backbone_unet3d import UNet3D


torch.set_float32_matmul_precision("medium")


@hydra.main(version_base="1.2", config_path=".", config_name="segmentation_config")
def main(cfg: DictConfig) -> None:
    cfg = hydra.utils.instantiate(cfg)
    results_path = cfg.results_dir
    os.environ["CUDA_VISIBLE_DEVICES"] = ",".join([str(x) for x in cfg.cuda_devices])

    assert (
        sum(
            cfg[weights] is not None
            for weights in [
                "suprem_weights",
                "dino_weights",
                "resume_training_from",
                "test_ckpt",
            ]
        )
        <= 1
    ), "only one weights/ckpt path can be used at a time"

    model_dir = os.path.join(results_path, cfg.experiment)
    if not os.path.isdir(model_dir):
        os.makedirs(model_dir)

    tags = cfg.tags + [cfg.backbone]
    print(f"EXPERIMENT : {cfg.experiment}")
    if cfg.clearml:
        if len(glob(model_dir + "/*.cmlid")) == 0:
            task = Task.create(
                project_name=cfg.clearml_project_name, task_name=cfg.experiment
            )
            with open(os.path.join(model_dir, f"{task.id}.cmlid"), "w"):
                pass

        clearml_task_id = glob(model_dir + "/*.cmlid")[0].split("/")[-1].split(".")[0]
        task = Task.init(
            project_name=cfg.clearml_project_name,
            task_name=cfg.experiment,
            reuse_last_task_id=clearml_task_id,
            continue_last_task=0,
            tags=tags,
        )

        task.connect(cfg)
        task.add_tags(cfg.tags)
    # Model definition:
    ##############################################################################
    if cfg.backbone == "unet3d":
        backbone = UNet3D(for_cls=False)

        if cfg.dino_weights is not None:
            print(f"LOADING ckpt from {cfg.dino_weights}")
            state_dict = torch.load(
                open(cfg.dino_weights, "rb"), map_location=torch.device("cpu")
            )
            state_dict = {
                k.replace("teacher_backbone.", ""): v
                for k, v in state_dict["state_dict"].items()
                if "teacher_backbone." in k
            }
        elif cfg.suprem_weights is not None:
            state_dict = torch.load(
                cfg.suprem_weights, map_location=torch.device("cpu")
            )
            state_dict = {
                k.replace("module.backbone.", ""): v
                for k, v in state_dict["net"].items()
            }
        else:
            state_dict = backbone.state_dict()

        backbone.load_state_dict(state_dict, strict=False)
        conv_inputs = [("model.backbone_features", 512)]

    heads = [
        HeadDenseSegmentation(
            head_name="head_seg",
            conv_inputs=conv_inputs,
            shared_classifier_head=nn.Conv3d(64, cfg.num_classes, kernel_size=1),
        )
    ]

    model = ModelMultiHead(conv_inputs=(("img", 1),), backbone=backbone, heads=heads)

    # read environment variables for data, cache and results locations

    ## Basic settings:
    ##############################################################################
    # create model results dir:
    # we use a time stamp in model directory name, to prevent re-writing

    # start logger
    fuse_logger_start(output_path=model_dir, console_verbose_level=logging.INFO)
    print("Done")

    # set constant seed for reproducibility.
    os.environ[
        "CUBLAS_WORKSPACE_CONFIG"
    ] = ":4096:8"  # required for pytorch deterministic mode
    rand_gen = Seed.set_seed(1234, deterministic_mode=True)

    ## FuseMedML dataset preparation
    ##############################################################################
    df_all = pd.read_csv(cfg.csv_path)
    dfs = {}
    for _set in ["train", "val", "test"]:
        dfs[_set] = df_all[df_all.fold.isin(cfg[f"{_set}_folds"])]

    train_ds = SegOAI.dataset(
        dfs["train"], validation=(not cfg.aug), resize_to=cfg.resize_to
    )
    val_ds = SegOAI.dataset(dfs["val"], validation=True, resize_to=cfg.resize_to)
    test_ds = SegOAI.dataset(dfs["test"], validation=True, resize_to=cfg.resize_to)
    ## Create dataloader

    train_dl = DataLoader(
        dataset=train_ds,
        shuffle=True,
        drop_last=False,
        # batch_sampler=sampler,
        batch_size=cfg.batch_size,
        collate_fn=CollateDefault(),
        num_workers=cfg.n_workers,
    )

    val_dl = DataLoader(
        dataset=val_ds,
        shuffle=False,
        drop_last=False,
        batch_sampler=None,
        batch_size=1,
        num_workers=cfg.n_workers,
        collate_fn=CollateDefault(),
        generator=rand_gen,
    )
    test_dl = DataLoader(
        dataset=test_ds,
        shuffle=False,
        drop_last=False,
        batch_sampler=None,
        batch_size=1,
        num_workers=cfg.n_workers,
        collate_fn=CollateDefault(),
        generator=rand_gen,
    )

    # Loss definition:
    ##############################################################################
    callable_loss = GeneralizedDiceLoss(
        include_background=cfg.include_background,
        sigmoid=cfg.sigmoid,
        softmax=cfg.softmax,
    )
    losses = {
        "generalized_dice_loss": LossDefault(
            pred="model.logits.head_seg",
            target="seg",
            callable=callable_loss,
            weight=1.0,
        )
    }

    # Metrics definition:
    ##############################################################################
    def pre_process_for_dice(sample: NDict) -> NDict:
        sample["model.logits.head_seg"] = sample["model.logits.head_seg"].argmax(axis=0)
        sample["seg"] = sample["seg"].argmax(axis=0)
        return sample

    train_metrics = {}
    val_metrics = {
        "dice": MetricDice(
            pred="model.logits.head_seg",
            target="seg",
            pre_collect_process_func=pre_process_for_dice,
        )
    }

    # best_epoch_source = dict(
    #     monitor="validation.metrics.auc.macro_avg",  # can be any key from losses or metrics dictionaries
    #     mode="max",  # can be either min/max
    # )

    # Optimizer definition:
    ##############################################################################

    optimizer = optim.Adam(
        model.parameters(), lr=cfg.learning_rate, weight_decay=cfg.weight_decay
    )

    # Scheduler definition:
    ##############################################################################
    # lr_scheduler = cosine_annealing_with_warmup_lr_scheduler(optimizer, num_warmup_steps=cfg.num_warmup_steps, T_max=cfg.n_epochs, eta_min=0.00001)
    lr_scheduler = optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max=cfg.n_epochs)

    lr_sch_config = dict(scheduler=lr_scheduler)
    optimizers_and_lr_schs = dict(optimizer=optimizer)
    optimizers_and_lr_schs["lr_scheduler"] = lr_sch_config

    # CallBacks

    callbacks = [
        # BackboneFinetuning(unfreeze_backbone_at_epoch=8,),
        LearningRateMonitor(logging_interval="epoch"),
        # ModelCheckpoint(dirpath=model_dir, save_last=True)
    ]

    ## Training
    ##############################################################################
    ckpt_path = None
    if cfg.resume_training_from is not None:
        ckpt_path = cfg.ckpt_path

    # create instance of PL module - FuseMedML generic version
    pl_module = LightningModuleDefault(
        model_dir=model_dir,
        model=model,
        losses=losses,
        train_metrics=train_metrics,
        validation_metrics=val_metrics,
        test_metrics=val_metrics,
        # best_epoch_source=best_epoch_source,
        optimizers_and_lr_schs=optimizers_and_lr_schs,
        callbacks=callbacks,
        save_model=False,
        log_unit="epoch",
    )
    # create lightining trainer.
    devices = torch.cuda.device_count()
    pl_trainer = pl.Trainer(
        default_root_dir=model_dir,
        max_epochs=cfg.n_epochs,
        accelerator="gpu",
        devices=devices,
        strategy="ddp" if devices > 1 else "auto",
        num_sanity_val_steps=0,
        gradient_clip_val=cfg.grad_clip,
        deterministic=False,
        accumulate_grad_batches=cfg.accumulate_grad_batches,
        precision=cfg.precision,
    )
    if cfg.test_ckpt is not None:
        print(f"Test using ckpt: {cfg.test_ckpt}")
        pl_trainer.validate(pl_module, test_dl, ckpt_path=cfg.test_ckpt)
    else:
        # print(f"Training using ckpt: {ckpt_path}")
        pl_trainer.fit(pl_module, train_dl, val_dl, ckpt_path=ckpt_path)


if __name__ == "__main__":
    main()



from fuse.dl.cross_validation.pipeline import run
from fuse_examples.imaging.classification.stoic21.runner_stoic21 import (
    run_train,
    run_infer,
    run_eval,
)
from fuse_examples.imaging.classification.stoic21.dataset import (
    create_dataset,
    train_val_test_splits,
)
from fuse_examples.imaging.classification.stoic21.runner_stoic21 import (
    TRAIN_COMMON_PARAMS,
    INFER_COMMON_PARAMS,
    EVAL_COMMON_PARAMS,
    DATASET_COMMON_PARAMS,
)
import os

##########################################
# Required Parameters
##########################################
num_repetitions = 1
num_gpus_total = 3
num_gpus_per_split = 1
num_folds = 4
num_folds_used = 4
dataset_func = create_dataset
train_func = run_train
infer_func = run_infer
eval_func = run_eval
deterministic_mode = True

# output paths:
root_path = "results"
paths = {
    "data_dir": os.environ["STOIC21_DATA_PATH"],
    "cache_dir": os.path.join(root_path, "stoic21/cache_dir"),
    "model_dir": os.path.join(root_path, "stoic21/model_dir"),
    "data_split_filename": os.path.join(root_path, "stoic21/stoic21_split.pkl"),
    "inference_dir": os.path.join(root_path, "stoic21/infer_dir"),
    "eval_dir": os.path.join(root_path, "stoic21/eval_dir"),
}


##########################################
# Custom Parameters
##########################################


##########################################
# Train Params
##########################################
train_params = TRAIN_COMMON_PARAMS
train_params["train_func"] = run_train
train_params["trainer.num_devices"] = num_gpus_per_split
train_params["trainer.strategy"] = None
train_params["trainer.auto_select_gpus"] = False
######################################
# Inference Params
######################################
infer_params = INFER_COMMON_PARAMS
infer_params["run_func"] = run_infer
infer_params["infer_filename"] = "infer.gz"
infer_params["trainer.auto_select_gpus"] = False
infer_params["pred_key"] = "model.output.classification"

######################################
# Eval Params
######################################
eval_params = EVAL_COMMON_PARAMS
eval_params["infer_filename"] = infer_params["infer_filename"]
eval_params["run_func"] = run_eval

##########################################
# Dataset Params
##########################################

dataset_params = DATASET_COMMON_PARAMS
dataset_params["train"] = train_params
dataset_params["infer"] = infer_params
dataset_params["target_key"] = "data.gt.probSevere"

splits = train_val_test_splits(paths=paths, params=dataset_params)
sample_ids_per_fold = [(s[0], s[1]) for s in splits]

if __name__ == "__main__":
    run(
        num_folds,
        num_folds_used,
        num_gpus_total,
        num_gpus_per_split,
        num_repetitions,
        dataset_func,
        train_func,
        infer_func,
        eval_func,
        dataset_params,
        train_params,
        infer_params,
        eval_params,
        paths,
        deterministic_mode,
        sample_ids_per_fold,
    )



"""
(C) Copyright 2021 IBM Corp.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Created on June 30, 2021

"""

import os
import copy
import logging
from typing import OrderedDict, Sequence, Tuple

import torch
import torch.optim as optim
import torch.nn.functional as F
from torch.utils.data.dataloader import DataLoader
from torchvision.models.resnet import ResNet18_Weights

from fuse.dl.models import ModelMultiHead
from fuse.dl.models.backbones.backbone_resnet import BackboneResnet
from fuse.dl.models.heads.head_global_pooling_classifier import (
    HeadGlobalPoolingClassifier,
)
from fuse.dl.models.backbones.backbone_inception_resnet_v2 import (
    BackboneInceptionResnetV2,
)

from fuse.eval.metrics.classification.metrics_thresholding_common import (
    MetricApplyThresholds,
)
from fuse.eval.metrics.classification.metrics_classification_common import (
    MetricAccuracy,
    MetricAUCROC,
    MetricROCCurve,
)

from fuse.utils.utils_debug import FuseDebug
from fuse.utils.utils_logger import fuse_logger_start
from fuse.utils.file_io.file_io import create_dir, save_dataframe, load_pickle
import fuse.utils.gpu as GPU

from fuse.eval.evaluator import EvaluatorDefault
from fuse.dl.losses.loss_default import LossDefault
from fuse.data.utils.collates import CollateDefault

import pytorch_lightning as pl
from fuse.dl.lightning.pl_module import LightningModuleDefault
from fuse.dl.lightning.pl_funcs import convert_predictions_to_dataframe
from pytorch_lightning.utilities.rank_zero import rank_zero_only

from fuseimg.datasets.isic import ISIC, ISICDataModule
from fuse_examples.imaging.classification.isic.golden_members import FULL_GOLDEN_MEMBERS

import torch.nn as nn
from fuse.dl.models.model_wrapper import ModelWrapSeqToDict
from fuse.dl.models.backbones.backbone_vit import ViT


###########################################################################################################
# Fuse
###########################################################################################################
##########################################
# Debug modes
##########################################
mode = "default"  # Options: 'default', 'debug'. See details in FuseDebug
debug = FuseDebug(mode)

##########################################
# GPUs and Workers
##########################################
NUM_GPUS = 1  # supports multiple gpu training with DDP strategy
NUM_WORKERS = 8

##########################################
# Modality
##########################################
multimodality = (
    True  # Set: 'False' to use only imaging, 'True' to use imaging & meta-data
)

##########################################
# Model Type
##########################################
model_type = "Transformer"  # Set: 'Transformer' to use ViT/MMViT, 'CNN' to use InceptionResNet or Resnet18

##########################################
# Output Paths
##########################################


ROOT = "./_examples/isic/"
DATA = (
    os.environ["ISIC19_DATA_PATH"]
    if "ISIC19_DATA_PATH" in os.environ
    else os.path.join(ROOT, "data_dir")
)
modality = "multimodality" if multimodality else "imaging"
model_dir = os.path.join(ROOT, f"model_dir_{modality}")
PATHS = {
    "model_dir": model_dir,
    "inference_dir": os.path.join(model_dir, "infer_dir"),
    "eval_dir": os.path.join(model_dir, "eval_dir"),
    "data_dir": DATA,
    "cache_dir": os.path.join(ROOT, "cache_dir"),
    "data_split_filename": os.path.join(model_dir, "isic_split.pkl"),
}


##########################################
# Train Common Params
##########################################
TRAIN_COMMON_PARAMS = {}
# ============
# Data
# ============
TRAIN_COMMON_PARAMS[
    "data.batch_size"
] = 32  # effective batch size = batch_size * num_gpus
TRAIN_COMMON_PARAMS["data.num_workers"] = NUM_WORKERS
TRAIN_COMMON_PARAMS["data.num_folds"] = 5
TRAIN_COMMON_PARAMS["data.train_folds"] = [0, 1, 2]
TRAIN_COMMON_PARAMS["data.validation_folds"] = [3]
TRAIN_COMMON_PARAMS["data.infer_folds"] = [4]
TRAIN_COMMON_PARAMS["data.samples_ids"] = {"all": None, "golden": FULL_GOLDEN_MEMBERS}[
    "all"
]


# ===============
# PL Trainer
# ===============
TRAIN_COMMON_PARAMS["trainer.num_epochs"] = 30
TRAIN_COMMON_PARAMS["trainer.num_devices"] = NUM_GPUS
TRAIN_COMMON_PARAMS["trainer.accelerator"] = "gpu"

# ===============
# Optimizer
# ===============
TRAIN_COMMON_PARAMS["opt.lr"] = 1e-5
TRAIN_COMMON_PARAMS["opt.weight_decay"] = 1e-3

# ===============
# Model
# ===============
if model_type == "CNN":
    TRAIN_COMMON_PARAMS["model"] = dict(
        dropout_rate=0.5,
        layers_description=(256,),
        tabular_data_inputs=[("data.input.clinical.all", 19)]
        if multimodality
        else None,
        tabular_layers_description=(128,) if multimodality else tuple(),
    )
elif model_type == "Transformer":
    token_dim = 768
    TRAIN_COMMON_PARAMS["model"] = dict(
        token_dim=token_dim,
        projection_kwargs=dict(
            image_shape=[300, 300], patch_shape=[30, 30], channels=3
        ),
        transformer_kwargs=dict(
            depth=12,
            heads=12,
            mlp_dim=token_dim * 4,
            dim_head=64,
            dropout=0.0,
            emb_dropout=0.0,
        ),
    )

TRAIN_COMMON_PARAMS["model_type"] = model_type


def perform_softmax(logits: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
    cls_preds = F.softmax(logits, dim=1)
    return logits, cls_preds


class MMViT(ViT):
    def __init__(
        self,
        token_dim: int,
        projection_kwargs: dict,
        transformer_kwargs: dict,
        multimodality: bool,
    ):
        super().__init__(token_dim, projection_kwargs, transformer_kwargs)
        self.multimodality = multimodality
        num_tokens = self.projection_layer.num_tokens
        self.token_dim = token_dim
        self._head = nn.Linear(token_dim, 8)
        if self.multimodality:
            # change pos embedding to accept additional token for multimodal
            self.transformer.pos_embedding = nn.Parameter(
                torch.randn(1, num_tokens + 2, token_dim)
            )

    # This forward can be Multimodal or just Imaging
    def forward(
        self, img_x: torch.Tensor, clinical_x: torch.Tensor = None
    ) -> torch.Tensor:
        img_x = self.projection_layer(img_x)
        if self.multimodality:
            clinical_x = clinical_x.unsqueeze(1)
            clinical_x_zeros = torch.zeros((img_x.shape[0], 1, self.token_dim))
            clinical_x_zeros[:, :, :19] = clinical_x
            clinical_x = clinical_x_zeros.cuda()
            x = torch.cat((img_x, clinical_x), 1)
        else:
            x = img_x
        x = self.transformer(x)
        x = self._head(x[:, 0])
        return x


def create_transformer_model(
    token_dim: int,
    projection_kwargs: dict,
    transformer_kwargs: dict,
) -> ModelWrapSeqToDict:
    torch_model = MMViT(
        token_dim=token_dim,
        projection_kwargs=projection_kwargs,
        transformer_kwargs=transformer_kwargs,
        multimodality=multimodality,
    )
    model = ModelWrapSeqToDict(
        model=torch_model,
        model_inputs=["data.input.img", "data.input.clinical.all"]
        if multimodality
        else ["data.input.img"],
        post_forward_processing_function=perform_softmax,
        model_outputs=["model.logits.head_0", "model.output.head_0"],
    )
    return model


def create_cnn_model(
    dropout_rate: float,
    layers_description: Sequence[int],
    tabular_data_inputs: Sequence[Tuple[str, int]],
    tabular_layers_description: Sequence[int],
    backbone_type: str = "Resnet18",
) -> torch.nn.Module:
    """
    creates the model

    :param backbone_type: (str) "InceptionResnetV2" or "Resnet18"
    """
    if backbone_type == "Resnet18":
        backbone = BackboneResnet(
            weights=ResNet18_Weights.IMAGENET1K_V1, in_channels=3, name="resnet18"
        )
        header_conv_inputs = [("model.backbone_features", 512)]
    elif backbone_type == "InceptionResnetV2":
        backbone = BackboneInceptionResnetV2(input_channels_num=3, logical_units_num=43)
        header_conv_inputs = [("model.backbone_features", 1536)]
    else:
        raise Exception(f"backbone_type ({backbone_type}) not supported")

    model = ModelMultiHead(
        conv_inputs=(("data.input.img", 3),),
        backbone=backbone,
        heads=[
            HeadGlobalPoolingClassifier(
                head_name="head_0",
                dropout_rate=dropout_rate,
                conv_inputs=header_conv_inputs,
                tabular_data_inputs=tabular_data_inputs,
                layers_description=layers_description,
                tabular_layers_description=tabular_layers_description,
                num_classes=8,
                pooling="avg",
            ),
        ],
    )
    return model


def create_datamodule(paths: dict, train_common_params: dict) -> pl.LightningDataModule:
    """
    In order to support the DDP strategy one need to create a Lightning Data Module.
    """
    datamodule = ISICDataModule(
        data_dir=paths["data_dir"],
        cache_dir=paths["cache_dir"],
        num_workers=train_common_params["data.num_workers"],
        batch_size=train_common_params["data.batch_size"],
        train_folds=train_common_params["data.train_folds"],
        validation_folds=train_common_params["data.validation_folds"],
        infer_folds=train_common_params["data.infer_folds"],
        split_filename=paths["data_split_filename"],
        sample_ids=train_common_params["data.samples_ids"],
        reset_cache=False,
        reset_split=False,
        use_batch_sampler=True if NUM_GPUS <= 1 else False,
        verify_folds_total_size=True,
    )

    return datamodule


#################################
# Train Template
#################################
def run_train(paths: dict, train_common_params: dict) -> None:
    # ==============================================================================
    # Logger
    # ==============================================================================
    fuse_logger_start(
        output_path=paths["model_dir"], console_verbose_level=logging.INFO
    )
    lgr = logging.getLogger("Fuse")
    lgr.info("Fuse Train", {"attrs": ["bold", "underline"]})

    lgr.info(f'model_dir={paths["model_dir"]}', {"color": "magenta"})
    lgr.info(f'data_dir={paths["data_dir"]}', {"color": "magenta"})
    lgr.info(f'cache_dir={paths["cache_dir"]}', {"color": "magenta"})

    # ==============================================================================
    # Data
    # ==============================================================================
    lgr.info("Datamodule:", {"attrs": "bold"})

    datamodule = create_datamodule(paths, train_common_params)

    lgr.info("Datamodule: Done", {"attrs": "bold"})

    # ==============================================================================
    # Model
    # ==============================================================================
    lgr.info("Model:", {"attrs": "bold"})
    model_type = train_common_params["model_type"]
    if model_type == "Transformer":
        model = create_transformer_model(**train_common_params["model"])
    elif model_type == "CNN":
        model = create_cnn_model(**train_common_params["model"])

    lgr.info("Model: Done", {"attrs": "bold"})

    # ====================================================================================
    #  Loss
    # ====================================================================================
    losses = {
        "cls_loss": LossDefault(
            pred="model.logits.head_0",
            target="data.label",
            callable=F.cross_entropy,
            weight=1.0,
        ),
    }

    # ====================================================================================
    # Metrics
    # ====================================================================================
    class_names = ["MEL", "NV", "BCC", "AK", "BKL", "DF", "VASC", "SCC"]
    train_metrics = OrderedDict(
        [
            (
                "op",
                MetricApplyThresholds(pred="model.output.head_0"),
            ),  # will apply argmax
            (
                "auc",
                MetricAUCROC(
                    pred="model.output.head_0",
                    target="data.label",
                    class_names=class_names,
                ),
            ),
            (
                "accuracy",
                MetricAccuracy(pred="results:metrics.op.cls_pred", target="data.label"),
            ),
        ]
    )

    validation_metrics = copy.deepcopy(
        train_metrics
    )  # use the same metrics in validation as well

    best_epoch_source = dict(monitor="validation.metrics.auc.macro_avg", mode="max")

    # create optimizer
    optimizer = optim.Adam(
        model.parameters(),
        lr=train_common_params["opt.lr"],
        weight_decay=train_common_params["opt.weight_decay"],
    )

    # create learning scheduler
    lr_scheduler = {
        "ReduceLROnPlateau": optim.lr_scheduler.ReduceLROnPlateau(optimizer),
        "CosineAnnealing": optim.lr_scheduler.CosineAnnealingWarmRestarts(
            optimizer, T_0=1
        ),
    }["ReduceLROnPlateau"]
    lr_sch_config = dict(scheduler=lr_scheduler, monitor="validation.losses.total_loss")

    # optimizer and lr sch - see pl.LightningModule.configure_optimizers return value for all options
    optimizers_and_lr_schs = dict(optimizer=optimizer, lr_scheduler=lr_sch_config)

    # =====================================================================================
    #  Train
    # =====================================================================================
    lgr.info("Train:", {"attrs": "bold"})

    # create instance of PL module - FuseMedML generic version
    pl_module = LightningModuleDefault(
        model_dir=paths["model_dir"],
        model=model,
        losses=losses,
        train_metrics=train_metrics,
        validation_metrics=validation_metrics,
        best_epoch_source=best_epoch_source,
        optimizers_and_lr_schs=optimizers_and_lr_schs,
    )

    # create lightning trainer.
    pl_trainer = pl.Trainer(
        default_root_dir=paths["model_dir"],
        max_epochs=train_common_params["trainer.num_epochs"],
        accelerator=train_common_params["trainer.accelerator"],
        devices=train_common_params["trainer.num_devices"],
    )

    # train
    pl_trainer.fit(pl_module, datamodule=datamodule)

    lgr.info("Train: Done", {"attrs": "bold"})


######################################
# Inference Common Params
######################################
INFER_COMMON_PARAMS = {}
INFER_COMMON_PARAMS["infer_filename"] = "infer_file.gz"
INFER_COMMON_PARAMS["checkpoint"] = "best_epoch.ckpt"
INFER_COMMON_PARAMS["data.num_workers"] = NUM_WORKERS
INFER_COMMON_PARAMS["data.infer_folds"] = [4]  # infer validation set
INFER_COMMON_PARAMS["data.batch_size"] = 4

INFER_COMMON_PARAMS["model"] = TRAIN_COMMON_PARAMS["model"]
INFER_COMMON_PARAMS["trainer.num_devices"] = 1  # No need for multi-gpu in inference
INFER_COMMON_PARAMS["trainer.accelerator"] = TRAIN_COMMON_PARAMS["trainer.accelerator"]
INFER_COMMON_PARAMS["model_type"] = TRAIN_COMMON_PARAMS["model_type"]

######################################
# Inference Template
######################################


@rank_zero_only
def run_infer(paths: dict, infer_common_params: dict) -> None:
    create_dir(paths["inference_dir"])
    infer_file = os.path.join(
        paths["inference_dir"], infer_common_params["infer_filename"]
    )
    checkpoint_file = os.path.join(
        paths["model_dir"], infer_common_params["checkpoint"]
    )

    ## Logger
    fuse_logger_start(
        output_path=paths["inference_dir"], console_verbose_level=logging.INFO
    )
    lgr = logging.getLogger("Fuse")
    lgr.info("Fuse Inference", {"attrs": ["bold", "underline"]})
    lgr.info(f"infer_filename={infer_file}", {"color": "magenta"})

    ## Data
    folds = load_pickle(
        paths["data_split_filename"]
    )  # assume exists and created in train func

    infer_sample_ids = []
    for fold in infer_common_params["data.infer_folds"]:
        infer_sample_ids += folds[fold]

    # Create dataset
    infer_dataset = ISIC.dataset(
        paths["data_dir"], paths["cache_dir"], samples_ids=infer_sample_ids, train=False
    )

    # dataloader
    infer_dataloader = DataLoader(
        dataset=infer_dataset,
        collate_fn=CollateDefault(),
        batch_size=infer_common_params["data.batch_size"],
        num_workers=infer_common_params["data.num_workers"],
    )

    # load python lightning module
    model_type = infer_common_params["model_type"]
    if model_type == "Transformer":
        model = create_transformer_model(**infer_common_params["model"])
    elif model_type == "CNN":
        model = create_cnn_model(**infer_common_params["model"])

    pl_module = LightningModuleDefault.load_from_checkpoint(
        checkpoint_file,
        model_dir=paths["model_dir"],
        model=model,
        map_location="cpu",
        strict=True,
    )
    # set the prediction keys to extract (the ones used be the evaluation function).
    pl_module.set_predictions_keys(
        ["model.output.head_0", "data.label"]
    )  # which keys to extract and dump into file

    # create a trainer instance
    pl_trainer = pl.Trainer(
        default_root_dir=paths["model_dir"],
        accelerator=infer_common_params["trainer.accelerator"],
        devices=infer_common_params["trainer.num_devices"],
        max_epochs=0,
    )
    predictions = pl_trainer.predict(
        pl_module, infer_dataloader, return_predictions=True
    )

    # convert list of batch outputs into a dataframe
    infer_df = convert_predictions_to_dataframe(predictions)
    save_dataframe(infer_df, infer_file)


######################################
# Eval Common Params
######################################
EVAL_COMMON_PARAMS = {}
EVAL_COMMON_PARAMS["infer_filename"] = INFER_COMMON_PARAMS["infer_filename"]


######################################
# Eval Template
######################################
@rank_zero_only
def run_eval(paths: dict, eval_common_params: dict) -> None:
    infer_file = os.path.join(
        paths["inference_dir"], eval_common_params["infer_filename"]
    )

    fuse_logger_start(output_path=None, console_verbose_level=logging.INFO)
    lgr = logging.getLogger("Fuse")
    lgr.info("Fuse Eval", {"attrs": ["bold", "underline"]})

    # metrics
    metrics = OrderedDict(
        [
            (
                "op",
                MetricApplyThresholds(pred="model.output.head_0"),
            ),  # will apply argmax
            ("auc", MetricAUCROC(pred="model.output.head_0", target="data.label")),
            (
                "accuracy",
                MetricAccuracy(pred="results:metrics.op.cls_pred", target="data.label"),
            ),
            (
                "roc",
                MetricROCCurve(
                    pred="model.output.head_0",
                    target="data.label",
                    output_filename=os.path.join(
                        paths["inference_dir"], "roc_curve.png"
                    ),
                ),
            ),
        ]
    )

    # create evaluator
    evaluator = EvaluatorDefault()

    # run
    results = evaluator.eval(
        ids=None, data=infer_file, metrics=metrics, output_dir=paths["eval_dir"]
    )

    return results


######################################
# Run
######################################
if __name__ == "__main__":
    ## allocate gpus
    # uncomment if you want to use specific gpus instead of automatically looking for free ones
    force_gpus = None  # [0]
    GPU.choose_and_enable_multiple_gpus(NUM_GPUS, force_gpus=force_gpus)

    ISIC.download(data_path=PATHS["data_dir"])

    RUNNING_MODES = ["train", "infer", "eval"]  # Options: 'train', 'infer', 'eval'

    # train
    if "train" in RUNNING_MODES:
        run_train(paths=PATHS, train_common_params=TRAIN_COMMON_PARAMS)

    # infer
    if "infer" in RUNNING_MODES:
        run_infer(paths=PATHS, infer_common_params=INFER_COMMON_PARAMS)

    # eval
    if "eval" in RUNNING_MODES:
        run_eval(paths=PATHS, eval_common_params=EVAL_COMMON_PARAMS)
